# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Lint as: python3
"""Tests for the Shapenet dataset."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import tensorflow_datasets as tfds
from tensorflow_graphics.datasets.shapenet import shapenet

_EXAMPLE_DIR = os.path.join(os.path.dirname(__file__), 'fakes')


class ShapenetTest(tfds.testing.DatasetBuilderTestCase):
  DATASET_CLASS = shapenet.Shapenet
  SPLITS = {
      'train': 3,  # Number of fake train example
      'test': 1,  # Number of fake test example
      'validation': 1  # Number of fake validation examples
  }

  DL_EXTRACT_RESULT = 'all.csv'
  EXAMPLE_DIR = _EXAMPLE_DIR

  EXPECTED_ITEMS = {
      'train': {
          b'a98038807a61926abce962d6c4b37336': {
              'synset': '02691156',
              'num_vertices': 8,
              'num_faces': 12
          },
          b'a800bd725fe116447a84e76181a9e08f': {
              'synset': '03001627',
              'num_vertices': 6,
              'num_faces': 9
          },
          b'9550774ad1c19b24a5a118bd15e6e34f': {
              'synset': '02691156',
              'num_vertices': 12,
              'num_faces': 20
          },
      },
      'test': {
          b'3d5354863690ac7eca27bba175814d1': {
              'synset': '02691156',
              'num_vertices': 4,
              'num_faces': 4
          }
      },
      'validation': {
          b'7eff60e0d72800b8ca8607f540cc62ba': {
              'synset': '02691156',
              'num_vertices': 6,
              'num_faces': 8
          },
      }
  }

  def test_dataset_items(self):
    builder = shapenet.Shapenet(data_dir=self.tmp_dir)
    self._download_and_prepare_as_dataset(builder)
    for split_name in self.SPLITS:
      items = tfds.as_numpy(builder.as_dataset(split=split_name))
      for item in items:
        expected = self.EXPECTED_ITEMS[split_name][item['model_id']]
        self.assertEqual(item['label'],
                         self._encode_synset(builder, expected['synset']))
        self.assertLen(item['trimesh']['vertices'], expected['num_vertices'])
        self.assertLen(item['trimesh']['faces'], expected['num_faces'])

  def _encode_synset(self, builder, synset):
    return builder.info.features['label'].encode_example(synset)


if __name__ == '__main__':
  tfds.testing.test_main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tensorflow.graphics B-spline interpolation module.

  This module supports cardinal B-spline interpolation up to degree 4, with up
  to C3 smoothness. It has functions to calculate basis functions, control point
  weights, and the final interpolation.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import enum
import tensorflow as tf

from tensorflow_graphics.util import asserts
from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import shape


class Degree(enum.IntEnum):
  """Defines valid degrees for B-spline interpolation."""
  CONSTANT = 0
  LINEAR = 1
  QUADRATIC = 2
  CUBIC = 3
  QUARTIC = 4


def _constant(position):
  """B-Spline basis function of degree 0 for positions in the range [0, 1]."""
  # A piecewise constant spline is discontinuous at the knots.
  return tf.expand_dims(tf.clip_by_value(1.0 + position, 1.0, 1.0), axis=-1)


def _linear(position):
  """B-Spline basis functions of degree 1 for positions in the range [0, 1]."""
  # Piecewise linear splines are C0 smooth.
  return tf.stack((1.0 - position, position), axis=-1)


def _quadratic(position):
  """B-Spline basis functions of degree 2 for positions in the range [0, 1]."""
  # We pre-calculate the terms that are used multiple times.
  pos_sq = tf.pow(position, 2.0)

  # Piecewise quadratic splines are C1 smooth.
  return tf.stack((tf.pow(1.0 - position, 2.0) / 2.0, -pos_sq + position + 0.5,
                   pos_sq / 2.0),
                  axis=-1)


def _cubic(position):
  """B-Spline basis functions of degree 3 for positions in the range [0, 1]."""
  # We pre-calculate the terms that are used multiple times.
  neg_pos = 1.0 - position
  pos_sq = tf.pow(position, 2.0)
  pos_cb = tf.pow(position, 3.0)

  # Piecewise cubic splines are C2 smooth.
  return tf.stack(
      (tf.pow(neg_pos, 3.0) / 6.0, (3.0 * pos_cb - 6.0 * pos_sq + 4.0) / 6.0,
       (-3.0 * pos_cb + 3.0 * pos_sq + 3.0 * position + 1.0) / 6.0,
       pos_cb / 6.0),
      axis=-1)


def _quartic(position):
  """B-Spline basis functions of degree 4 for positions in the range [0, 1]."""
  # We pre-calculate the terms that are used multiple times.
  neg_pos = 1.0 - position
  pos_sq = tf.pow(position, 2.0)
  pos_cb = tf.pow(position, 3.0)
  pos_qt = tf.pow(position, 4.0)

  # Piecewise quartic splines are C3 smooth.
  return tf.stack(
      (tf.pow(neg_pos, 4.0) / 24.0,
       (-4.0 * tf.pow(neg_pos, 4.0) + 4.0 * tf.pow(neg_pos, 3.0) +
        6.0 * tf.pow(neg_pos, 2.0) + 4.0 * neg_pos + 1.0) / 24.0,
       (pos_qt - 2.0 * pos_cb - pos_sq + 2.0 * position) / 4.0 + 11.0 / 24.0,
       (-4.0 * pos_qt + 4.0 * pos_cb + 6.0 * pos_sq + 4.0 * position + 1.0) /
       24.0, pos_qt / 24.0),
      axis=-1)


def knot_weights(positions,
                 num_knots,
                 degree,
                 cyclical,
                 sparse_mode=False,
                 name=None):
  """Function that converts cardinal B-spline positions to knot weights.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    positions: A tensor with shape `[A1, .. An]`. Positions must be between
      `[0, C - D)` for non-cyclical and `[0, C)` for cyclical splines, where `C`
      is the number of knots and `D` is the spline degree.
    num_knots: A strictly positive `int` describing the number of knots in the
      spline.
    degree: An `int` describing the degree of the spline, which must be smaller
      than `num_knots`.
    cyclical: A `bool` describing whether the spline is cyclical.
    sparse_mode: A `bool` describing whether to return a result only for the
      knots with nonzero weights. If set to True, the function returns the
      weights of only the `degree` + 1 knots that are non-zero, as well as the
      indices of the knots.
    name: A name for this op. Defaults to "bspline_knot_weights".

  Returns:
    A tensor with dense weights for each control point, with the shape
    `[A1, ... An, C]` if `sparse_mode` is False.
    Otherwise, returns a tensor of shape `[A1, ... An, D + 1]` that contains the
    non-zero weights, and a tensor with the indices of the knots, with the type
    tf.int32.

  Raises:
    ValueError: If degree is greater than 4 or num_knots - 1, or less than 0.
    InvalidArgumentError: If positions are not in the right range.
  """
  with tf.compat.v1.name_scope(name, "bspline_knot_weights", [positions]):
    positions = tf.convert_to_tensor(value=positions)

    if degree > 4 or degree < 0:
      raise ValueError("Degree should be between 0 and 4.")
    if degree > num_knots - 1:
      raise ValueError("Degree cannot be >= number of knots.")
    if cyclical:
      positions = asserts.assert_all_in_range(positions, 0.0, float(num_knots))
    else:
      positions = asserts.assert_all_in_range(positions, 0.0,
                                              float(num_knots - degree))

    all_basis_functions = {
        # Maps valid degrees to functions.
        Degree.CONSTANT: _constant,
        Degree.LINEAR: _linear,
        Degree.QUADRATIC: _quadratic,
        Degree.CUBIC: _cubic,
        Degree.QUARTIC: _quartic
    }
    basis_functions = all_basis_functions[degree]

    if not cyclical and num_knots - degree == 1:
      # In this case all weights are non-zero and we can just return them.
      if not sparse_mode:
        return basis_functions(positions)
      else:
        shift = tf.zeros_like(positions, dtype=tf.int32)
        return basis_functions(positions), shift

    # shape_batch = positions.shape.as_list()
    shape_batch = tf.shape(input=positions)
    positions = tf.reshape(positions, shape=(-1,))

    # Calculate the nonzero weights from the decimal parts of positions.
    shift = tf.floor(positions)
    sparse_weights = basis_functions(positions - shift)
    shift = tf.cast(shift, tf.int32)

    if sparse_mode:
      # Returns just the weights and the shift amounts, so that tf.gather_nd on
      # the knots can be used to sparsely activate knots if needed.
      shape_weights = tf.concat(
          (shape_batch, tf.constant((degree + 1,), dtype=tf.int32)), axis=0)
      sparse_weights = tf.reshape(sparse_weights, shape=shape_weights)
      shift = tf.reshape(shift, shape=shape_batch)
      return sparse_weights, shift

    num_positions = tf.size(input=positions)
    ind_row, ind_col = tf.meshgrid(
        tf.range(num_positions, dtype=tf.int32),
        tf.range(degree + 1, dtype=tf.int32),
        indexing="ij")

    tiled_shifts = tf.reshape(
        tf.tile(tf.expand_dims(shift, axis=-1), multiples=(1, degree + 1)),
        shape=(-1,))
    ind_col = tf.reshape(ind_col, shape=(-1,)) + tiled_shifts
    if cyclical:
      ind_col = tf.math.mod(ind_col, num_knots)
    indices = tf.stack((tf.reshape(ind_row, shape=(-1,)), ind_col), axis=-1)
    shape_indices = tf.concat((tf.reshape(
        num_positions, shape=(1,)), tf.constant(
            (degree + 1, 2), dtype=tf.int32)),
                              axis=0)
    indices = tf.reshape(indices, shape=shape_indices)
    shape_scatter = tf.concat((tf.reshape(
        num_positions, shape=(1,)), tf.constant((num_knots,), dtype=tf.int32)),
                              axis=0)
    weights = tf.scatter_nd(indices, sparse_weights, shape_scatter)
    shape_weights = tf.concat(
        (shape_batch, tf.constant((num_knots,), dtype=tf.int32)), axis=0)
    return tf.reshape(weights, shape=shape_weights)


def interpolate_with_weights(knots, weights, name=None):
  """Interpolates knots using knot weights.

  Note:
    In the following, A1 to An, and B1 to Bk are optional batch dimensions.

  Args:
    knots: A tensor with shape `[B1, ..., Bk, C]` containing knot values, where
      `C` is the number of knots.
    weights: A tensor with shape `[A1, ..., An, C]` containing dense weights for
      the knots, where `C` is the number of knots.
    name: A name for this op. Defaults to "bspline_interpolate_with_weights".

  Returns:
    A tensor with shape `[A1, ..., An, B1, ..., Bk]`, which is the result of
    spline interpolation.

  Raises:
    ValueError: If the last dimension of knots and weights is not equal.
  """
  with tf.compat.v1.name_scope(name, "bspline_interpolate_with_weights",
                               [knots, weights]):
    knots = tf.convert_to_tensor(value=knots)
    weights = tf.convert_to_tensor(value=weights)

    shape.compare_dimensions(
        tensors=(knots, weights), axes=-1, tensor_names=("knots", "weights"))

  return tf.tensordot(weights, knots, (-1, -1))


def interpolate(knots, positions, degree, cyclical, name=None):
  """Applies B-spline interpolation to input control points (knots).

  Note:
    In the following, A1 to An, and B1 to Bk are optional batch dimensions.

  Args:
    knots: A tensor with shape `[B1, ..., Bk, C]` containing knot values, where
      `C` is the number of knots.
    positions: Tensor with shape `[A1, .. An]`. Positions must be between `[0, C
      - D)` for non-cyclical and `[0, C)` for cyclical splines, where `C` is the
      number of knots and `D` is the spline degree.
    degree: An `int` between 0 and 4, or an enumerated constant from the Degree
      class, which is the degree of the splines.
    cyclical: A `bool`, whether the splines are cyclical.
    name: A name for this op. Defaults to "bspline_interpolate".

  Returns:
    A tensor of shape `[A1, ... An, B1, ..., Bk]`, which is the result of spline
    interpolation.
  """
  with tf.compat.v1.name_scope(name, "bspline_interpolate", [knots, positions]):
    knots = tf.convert_to_tensor(value=knots)
    positions = tf.convert_to_tensor(value=positions)

    num_knots = knots.get_shape().as_list()[-1]
    weights = knot_weights(positions, num_knots, degree, cyclical, False, name)
    return interpolate_with_weights(knots, weights)


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Lint as: python3
"""Triangle mesh feature."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import six
import tensorflow.compat.v2 as tf
from tensorflow_datasets import features
from tensorflow_graphics.io import triangle_mesh


class TriangleMesh(features.FeaturesDict):
  """`FeatureConnector` for triangle meshes.

  During `_generate_examples`, the feature connector accepts as input any of:

    * `str`: path to a {obj,stl,ply,glb} triangle mesh.
    * `trimesh.Trimesh`: A triangle mesh object.
    * `trimesh.Scene`: A scene object containing multiple TriangleMesh
       objects.
    * `dict:` A dictionary containing the vertices and faces of the mesh (see
       output format below).

  Output:
    A dictionary containing:
    # TODO(b/156112246): Add additional attributes (vertex normals, colors,
    # texture coordinates).

    * 'vertices': A `float32` tensor with shape `[N, 3]` denoting the vertex
    coordinates, where N is the number of vertices in the mesh.
    * 'faces': An `int64` tensor with shape `[F, 3]` denoting the face vertex
    indices, where F is the number of faces in the mesh.

    Note: In case the input specifies a Scene (with multiple meshes), the output
    will be a single TriangleMesh which combines all the triangle meshes in the
    scene.
  """

  def __init__(self):
    super(TriangleMesh, self).__init__({
        'vertices': features.Tensor(shape=(None, 3), dtype=tf.float32),
        'faces': features.Tensor(shape=(None, 3), dtype=tf.uint64),
    })

  def encode_example(self, path_or_trianglemesh):
    """Convert the given triangle mesh into a dict convertible to tf example."""
    if isinstance(path_or_trianglemesh, six.string_types):
      # The parameter is a path.
      with tf.io.gfile.GFile(path_or_trianglemesh, 'rb') as tmesh_file:
        features_dict = self._convert_to_trimesh_feature(
            triangle_mesh.load(tmesh_file))
    elif hasattr(path_or_trianglemesh, 'read') and hasattr(
        path_or_trianglemesh, 'name'):
      # The parameter is a file object.
      features_dict = self._convert_to_trimesh_feature(
          triangle_mesh.load(path_or_trianglemesh))
    elif isinstance(path_or_trianglemesh, dict):
      # The parameter is already a Trimesh dictionary.
      features_dict = path_or_trianglemesh
    else:
      # The parameter is a Trimesh or a Scene.
      features_dict = self._convert_to_trimesh_feature(path_or_trianglemesh)

    return super(TriangleMesh, self).encode_example(features_dict)

  def _convert_to_trimesh_feature(self, obj):
    if isinstance(obj, triangle_mesh.Trimesh):
      vertices = np.array(obj.vertices)
      faces = np.array(obj.faces, dtype=np.uint64)
    elif isinstance(obj, triangle_mesh.Scene):
      # Concatenate all the vertices and faces of the triangle meshes in the
      # scene.
      # TODO(b/156117488): Change to a different merging algorithm to avoid
      # duplicated vertices.
      vertices_list = [
          np.array(mesh.vertices) for mesh in obj.geometry.values()
      ]
      faces_list = np.array([
          np.array(mesh.faces, dtype=np.uint64)
          for mesh in obj.geometry.values()
      ])
      faces_offset = np.cumsum(
          [vertices.shape[0] for vertices in vertices_list], dtype=np.uint64)
      faces_list[1:] += faces_offset[:-1]
      vertices = np.concatenate(vertices_list, axis=0)
      faces = np.concatenate(faces_list, axis=0)
    else:
      raise ValueError('obj should be either a Trimesh or a Scene')
    return {
        'vertices': vertices.astype(np.float32),
        'faces': faces,
    }

  @classmethod
  def from_json_content(cls, value) -> 'TriangleMesh':
    return cls()

  def to_json_content(self):
    return {}



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements a differentiable rasterizer of triangular meshes.

The resulting rendering contains perspective-correct interpolation of attributes
defined at the vertices of the rasterized meshes. This rasterizer does not
provide gradients through visibility, but it does through visible geometry and
attributes.
"""

import tensorflow as tf

from tensorflow_graphics.rendering import rasterization_backend
from tensorflow_graphics.rendering.opengl import math as glm
from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import shape


def _perspective_correct_barycentrics(vertices_per_pixel, model_to_eye_matrix,
                                      perspective_matrix, image_size_float):
  """Creates the pixels grid and computes barycentrics."""
  # Construct the pixel grid with half-integer pixel centers.
  width = image_size_float[1]
  height = image_size_float[0]
  px = tf.linspace(0.5, width - 0.5, num=int(width))
  py = tf.linspace(0.5, height - 0.5, num=int(height))
  xv, yv = tf.meshgrid(px, py)
  pixel_position = tf.stack((xv, yv), axis=-1)

  return glm.perspective_correct_barycentrics(vertices_per_pixel,
                                              pixel_position,
                                              model_to_eye_matrix,
                                              perspective_matrix,
                                              (width, height))


def _perspective_correct_attributes(attribute, barycentrics, triangles,
                                    triangle_index, len_batch_shape):
  attribute = tf.gather(attribute, triangles, axis=-2)
  attribute_per_pixel = tf.gather(
      attribute, triangle_index, axis=-3, batch_dims=len_batch_shape)

  return glm.interpolate_attributes(attribute_per_pixel, barycentrics)


def _dim_value(dim):
  return 1 if dim is None else tf.compat.v1.dimension_value(dim)


def rasterize(vertices,
              triangles,
              attributes,
              model_to_eye_matrix,
              perspective_matrix,
              image_size,
              backend=rasterization_backend.RasterizationBackends.OPENGL,
              name=None):
  """Rasterizes the scene.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    vertices: A tensor of shape `[A1, ..., An, V, 3]` containing batches of `V`
      vertices, each defined by a 3D point.
    triangles: A tensor of shape `[T, 3]` containing `T` triangles, each
      associated with 3 vertices from `vertices`.
    attributes: A dictionary of tensors, each of shape `[A1, ..., An, V, K_a]`
      containing batches of `V` vertices, each associated with K-dimensional
      attributes. K_a may vary by attribute.
    model_to_eye_matrix: A tensor of shape `[A1, ..., An, 4, 4]` containing
      batches of matrices used to transform vertices from model to eye
      coordinates.
    perspective_matrix: A tensor of shape `[A1, ..., An, 4, 4]` containing
      batches of matrices used to project vertices from eye to clip coordinates.
    image_size: A tuple (height, width) containing the dimensions in pixels of
      the rasterized image.
    backend: A rasterization_backend.RasterizationBackends enum containing the
      backend method to use for rasterization.
    name: A name for this op. Defaults to 'triangle_rasterizer_rasterize'.

  Returns:
    A dictionary. The key "mask" is of shape `[A1, ..., An, height, width]` and
    stores a value of `0` of the pixel is assciated with the background, and `1`
    with the foreground. The key "barycentrics" is of shape
    `[A1, ..., An, height, width, 3]` and stores barycentric weights. Finally,
    the dictionary contains perspective correct interpolated attributes of shape
    `[A1, ..., An, height, width, K]` per entry in the `attributes` dictionary.
  """
  with tf.compat.v1.name_scope(name, "triangle_rasterizer_rasterize",
                               (vertices, triangles, attributes,
                                model_to_eye_matrix, perspective_matrix)):
    vertices = tf.convert_to_tensor(value=vertices)
    triangles = tf.convert_to_tensor(value=triangles)
    model_to_eye_matrix = tf.convert_to_tensor(value=model_to_eye_matrix)
    perspective_matrix = tf.convert_to_tensor(value=perspective_matrix)

    shape.check_static(
        tensor=vertices,
        tensor_name="vertices",
        has_rank_greater_than=1,
        has_dim_equals=((-1, 3)))
    shape.check_static(
        tensor=triangles,
        tensor_name="triangles",
        has_rank=2,
        has_dim_equals=((-1, 3)))
    shape.check_static(
        tensor=model_to_eye_matrix,
        tensor_name="model_to_eye_matrix",
        has_dim_equals=(((-2, 4), (-1, 4))))
    shape.check_static(
        tensor=perspective_matrix,
        tensor_name="perspective_matrix",
        has_dim_equals=(((-2, 4), (-1, 4))))

    image_size_float = (float(image_size[0]), float(image_size[1]))
    image_size_backend = (int(image_size[1]), int(image_size[0]))

    view_projection_matrix = tf.linalg.matmul(perspective_matrix,
                                              model_to_eye_matrix)
    triangle_index, _, mask = rasterization_backend.rasterize(
        vertices, triangles, view_projection_matrix, image_size_backend,
        backend)
    outputs = {"mask": mask, "triangle_indices": triangle_index}

    batch_shape = triangle_index.shape[:-3]
    batch_shape = [_dim_value(dim) for dim in batch_shape]

    vertices = tf.gather(vertices, triangles, axis=-2)

    # Gather does not work on negative indices, which is the case for the pixel
    # associated to the background.
    triangle_index = triangle_index * mask
    vertices_per_pixel = tf.gather(
        vertices, triangle_index, axis=-3, batch_dims=len(batch_shape))
    barycentrics = _perspective_correct_barycentrics(vertices_per_pixel,
                                                     model_to_eye_matrix,
                                                     perspective_matrix,
                                                     image_size_float)
    mask_float = tf.cast(tf.expand_dims(mask, axis=-1), vertices.dtype)
    outputs["barycentrics"] = mask_float * barycentrics

    for key, attribute in attributes.items():
      attribute = tf.convert_to_tensor(value=attribute)
      outputs[key] = mask_float * _perspective_correct_attributes(
          attribute, barycentrics, triangles, triangle_index, len(batch_shape))

    return outputs


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Lint as: python3
"""A thin wrapper around the trimesh library for loading triangle meshes."""

import os
import tensorflow as tf
import trimesh
from trimesh import Scene
from trimesh import Trimesh


# TODO(b/156115314): Revisit the library for loading the triangle meshes.
class GFileResolver(trimesh.visual.resolvers.Resolver):
  """A resolver using gfile for accessing other assets in the mesh directory."""

  def __init__(self, path):
    if tf.io.gfile.isdir(path):
      self.directory = path
    elif tf.io.gfile.exists(path):
      self.directory = os.path.dirname(path)
    else:
      raise ValueError('path is not a file or directory')

  def get(self, name):
    with tf.io.gfile.GFile(os.path.join(self.directory, name), 'rb') as f:
      data = f.read()
    return data


def load(file_obj, file_type=None, **kwargs):
  """Loads a triangle mesh from the given GFile/file path.

  Args:
    file_obj: A tf.io.gfile.GFile object or a string specifying the mesh file
      path.
    file_type: A string specifying the type of the file (e.g. 'obj', 'stl'). If
      not specified the file_type will be inferred from the file name.
    **kwargs: Additional arguments that should be passed to trimesh.load().

  Returns:
    A trimesh.Trimesh or trimesh.Scene.
  """

  if isinstance(file_obj, str):
    with tf.io.gfile.GFile(file_obj, 'r') as f:
      if file_type is None:
        file_type = trimesh.util.split_extension(file_obj)
      return trimesh.load(
          file_obj=f,
          file_type=file_type,
          resolver=GFileResolver(file_obj),
          **kwargs)

  if trimesh.util.is_file(file_obj):
    if not hasattr(file_obj, 'name') or not file_obj.name:
      raise ValueError(
          'file_obj must have attribute "name". Try passing the file name instead.'
      )
    if file_type is None:
      file_type = trimesh.util.split_extension(file_obj.name)
    return trimesh.load(
        file_obj=file_obj,
        file_type=file_type,
        resolver=GFileResolver(file_obj.name),
        **kwargs)

  raise ValueError('file_obj should be either a file object or a string')


__all__ = ['load', 'Trimesh', 'Scene']



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for bspline."""

from absl.testing import parameterized
import numpy as np

from tensorflow_graphics.math.interpolation import bspline
from tensorflow_graphics.util import test_case


class BSplineTest(test_case.TestCase):

  @parameterized.parameters((0.0, (1.0,)), (1.0, (1.0,)))
  def test_constant_basis_boundary_values(self, position, weights):
    """Tests that basis functions of degree 0 return expected values."""
    self.assertAllClose(bspline._constant(position), weights)  # pylint: disable=protected-access

  @parameterized.parameters((0.0, (1.0, 0.0)), (1.0, (0.0, 1.0)))
  def test_linear_basis_boundary_values(self, position, weights):
    """Tests that basis functions of degree 1 return expected values."""
    self.assertAllClose(bspline._linear(position), weights)  # pylint: disable=protected-access

  @parameterized.parameters((0.0, (0.5, 0.5, 0.0)), (1.0, (0.0, 0.5, 0.5)))
  def test_quadratic_basis_boundary_values(self, position, weights):
    """Tests that basis functions of degree 2 return expected values."""
    self.assertAllClose(bspline._quadratic(position), weights)  # pylint: disable=protected-access

  @parameterized.parameters((0.0, (1.0 / 6.0, 2.0 / 3.0, 1.0 / 6.0, 0.0)),
                            (1.0, (0.0, 1.0 / 6.0, 2.0 / 3.0, 1.0 / 6.0)))
  def test_cubic_basis_boundary_values(self, position, weights):
    """Tests that basis functions of degree 3 return expected values."""
    self.assertAllClose(bspline._cubic(position), weights)  # pylint: disable=protected-access

  @parameterized.parameters(
      (0.0, (1.0 / 24.0, 11.0 / 24.0, 11.0 / 24.0, 1.0 / 24.0, 0.0)),
      (1.0, (0.0, 1.0 / 24.0, 11.0 / 24.0, 11.0 / 24.0, 1.0 / 24.0)))
  def test_quartic_basis_boundary_values(self, position, weights):
    """Tests that basis functions of degree 4 return expected values."""
    self.assertAllClose(bspline._quartic(position), weights)  # pylint: disable=protected-access

  @parameterized.parameters(
      (((0.5,), (1.5,), (2.5,)), (((0.5, 0.5),), ((0.5, 0.5),), ((0.5, 0.5),)),
       (((0,), (1,), (2,))), 1, True),
      ((0.0, 1.0), ((0.5, 0.5, 0.0), (0.0, 0.5, 0.5)), (0, 0), 2, False),
  )
  def test_knot_weights_sparse_mode_preset(self, positions, gt_weights,
                                           gt_shifts, degree, cyclical):
    """Tests that sparse mode returns correct results."""
    weights, shifts = bspline.knot_weights(
        positions,
        num_knots=3,
        degree=degree,
        cyclical=cyclical,
        sparse_mode=True)

    self.assertAllClose(weights, gt_weights)
    self.assertAllClose(shifts, gt_shifts)

  @parameterized.parameters(
      (((0.5,),), (((0.5, 0.5, 0.0),),), 1),
      (((1.5,),), (((0.0, 0.5, 0.5),),), 1),
      (((2.5,),), (((0.5, 0.0, 0.5),),), 1),
      (((0.5,), (1.5,), (2.5,)),
       (((1.0 / 8.0, 0.75, 1.0 / 8.0),), ((1.0 / 8.0, 1.0 / 8.0, 0.75),),
        ((0.75, 1.0 / 8.0, 1.0 / 8.0),)), 2),
  )
  def test_knot_weights_preset(self, position, weights, degree):
    """Tests that knot weights are correct when degree < num_knots - 1."""
    self.assertAllClose(
        bspline.knot_weights(
            position, num_knots=3, degree=degree, cyclical=True), weights)

  @parameterized.parameters((((0.0,), (0.25,), (0.5,), (0.75,)),))
  def test_full_degree_non_cyclical_knot_weights(self, positions):
    """Tests that noncyclical weights are correct when using max degree."""
    cyclical_weights = bspline.knot_weights(
        positions=positions, num_knots=3, degree=2, cyclical=True)
    noncyclical_weights = bspline.knot_weights(
        positions=positions, num_knots=3, degree=2, cyclical=False)

    self.assertAllClose(cyclical_weights, noncyclical_weights)

  @parameterized.parameters(
      ("must have the same number of dimensions", ((None, 2), (None, 3, 3))),
      ("must have the same number of dimensions", ((2,), (3,))),
  )
  def test_interpolate_with_weights_exception_is_raised(self, error_msg,
                                                        shapes):
    """Tests that exception is raised when wrong number of knots is given."""
    self.assert_exception_is_raised(
        bspline.interpolate_with_weights, error_msg, shapes=shapes)

  @parameterized.parameters(
      (((0.5,), (0.0,), (0.9,)), (((0.5, 1.5), (1.5, 1.5), (2.5, 3.5)),)))
  def test_interpolate_with_weights_preset(self, positions, knots):
    """Tests that interpolate_with_weights works correctly."""
    degree = 1
    cyclical = False
    interp1 = bspline.interpolate(knots, positions, degree, cyclical)
    weights = bspline.knot_weights(positions, 2, degree, cyclical)
    interp2 = bspline.interpolate_with_weights(knots, weights)

    self.assertAllClose(interp1, interp2)

  @parameterized.parameters(
      (1, 2),
      (1, None),
      (2, 2),
      (2, None),
      (3, 2),
      (3, None),
      (4, 2),
      (4, None),
  )
  def test_knot_weights_exception_is_not_raised(self, positions_rank, dims):
    shapes = ([dims] * positions_rank,)

    self.assert_exception_is_not_raised(
        bspline.knot_weights,
        shapes=shapes,
        num_knots=3,
        degree=2,
        cyclical=True)

  @parameterized.parameters(
      ("Degree should be between 0 and 4.", 6, -1),
      ("Degree should be between 0 and 4.", 6, 5),
      ("Degree cannot be >= number of knots.", 2, 2),
      ("Degree cannot be >= number of knots.", 2, 3),
  )
  def test_knot_weights_exception_is_raised(self, error_msg, num_knots, degree):
    self.assert_exception_is_raised(
        bspline.knot_weights,
        error_msg,
        shapes=((10, 1),),
        num_knots=num_knots,
        degree=degree,
        cyclical=True)

  @parameterized.parameters(
      (1, 0, True),
      (1, 0, False),
      (2, 1, True),
      (2, 1, False),
      (3, 1, True),
      (3, 1, False),
      (3, 2, True),
      (3, 2, False),
      (4, 1, True),
      (4, 1, False),
      (4, 3, True),
      (4, 3, False),
      (5, 1, True),
      (5, 1, False),
      (5, 4, True),
      (5, 4, False),
  )
  def test_knot_weights_jacobian_is_correct(self, num_knots, degree, cyclical):
    """Tests that Jacobian is correct."""
    positions_init = np.random.random_sample((10, 1))
    scale = num_knots if cyclical else num_knots - degree
    positions_init *= scale

    def dense_mode_fn(positions):
      return bspline.knot_weights(
          positions=positions,
          num_knots=num_knots,
          degree=degree,
          cyclical=cyclical,
          sparse_mode=False)

    def sparse_mode_fn(positions):
      return bspline.knot_weights(
          positions=positions,
          num_knots=num_knots,
          degree=degree,
          cyclical=cyclical,
          sparse_mode=True)[0]

    with self.subTest(name="dense_mode"):
      self.assert_jacobian_is_correct_fn(dense_mode_fn, [positions_init])

    with self.subTest(name="sparse_mode"):
      self.assert_jacobian_is_correct_fn(sparse_mode_fn, [positions_init])


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tensorflow grid utility functions."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import shape


def _grid(starts, stops, nums):
  """Generates a M-D uniform axis-aligned grid.

  Warning:
    This op is not differentiable. Indeed, the gradient of tf.linspace and
    tf.meshgrid are currently not defined.

  Args:
    starts: A tensor of shape `[M]` representing the start points for each
      dimension.
    stops: A tensor of shape `[M]` representing the end points for each
      dimension.
    nums: A tensor of shape `[M]` representing the number of subdivisions for
      each dimension.

  Returns:
    A tensor of shape `[nums[0], ..., nums[M-1], M]` containing an M-D uniform
      grid.
  """
  params = [tf.unstack(tensor) for tensor in [starts, stops, nums]]
  layout = [tf.linspace(*param) for param in zip(*params)]
  return tf.stack(tf.meshgrid(*layout, indexing="ij"), axis=-1)


def generate(starts, stops, nums, name=None):
  r"""Generates a M-D uniform axis-aligned grid.

  Warning:
    This op is not differentiable. Indeed, the gradient of tf.linspace and
    tf.meshgrid are currently not defined.

  Note:
    In the following, `B` is an optional batch dimension.

  Args:
    starts: A tensor of shape `[M]` or `[B, M]`, where the last dimension
      represents a M-D start point.
    stops: A tensor of shape `[M]` or `[B, M]`, where the last dimension
      represents a M-D end point.
    nums: A tensor of shape `[M]` representing the number of subdivisions for
      each dimension.
    name: A name for this op. Defaults to "grid_generate".

  Returns:
    A tensor of shape `[nums[0], ..., nums[M-1], M]` containing an M-D uniform
      grid or a tensor of shape [B, nums[0], ..., nums[M-1], M]` containing B
      M-D uniform grids. Please refer to the example below for more details.

  Raises:
    ValueError: If the shape of `starts`, `stops`, or 'nums' is not supported.

  Examples:
    ```python
    print(generate((-1.0, -2.0), (1.0, 2.0), (3, 5)))
    >>> [[[-1. -2.]
          [-1. -1.]
          [-1.  0.]
          [-1.  1.]
          [-1.  2.]]
         [[ 0. -2.]
          [ 0. -1.]
          [ 0.  0.]
          [ 0.  1.]
          [ 0.  2.]]
         [[ 1. -2.]
          [ 1. -1.]
          [ 1.  0.]
          [ 1.  1.]
          [ 1.  2.]]]
    ```
    Generates a 3x5 2d grid from -1.0 to 1.0 with 3 subdivisions for the x
    axis and from -2.0 to 2.0 with 5 subdivisions for the y axis. This lead to a
    tensor of shape (3, 5, 2).
  """
  with tf.compat.v1.name_scope(name, "grid_generate", [starts, stops, nums]):
    starts = tf.convert_to_tensor(value=starts)
    stops = tf.convert_to_tensor(value=stops)
    nums = tf.convert_to_tensor(value=nums)

    shape.check_static(
        tensor=starts,
        tensor_name="starts",
        has_rank_greater_than=0,
        has_rank_less_than=3)
    shape.check_static(
        tensor=stops,
        tensor_name="stops",
        has_rank_greater_than=0,
        has_rank_less_than=3)
    shape.check_static(tensor=nums, tensor_name="nums", has_rank=1)
    shape.compare_batch_dimensions(
        tensors=(starts, stops), last_axes=(-1, -1), broadcast_compatible=False)
    shape.compare_dimensions((starts, stops, nums), -1,
                             ("starts", "stops", "nums"))

    if starts.shape.ndims == 1:
      return _grid(starts, stops, nums)
    else:
      return tf.stack([
          _grid(starts, stops, nums)
          for starts, stops in zip(tf.unstack(starts), tf.unstack(stops))
      ])


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Training loop for PointNet v1 on modelnet40."""
# pylint: disable=missing-function-docstring

import tensorflow as tf
from tensorflow_graphics.datasets import modelnet40
from tensorflow_graphics.nn.layer import pointnet
import tqdm  # pylint: disable=g-bad-import-order

from . import augment  # pylint: disable=g-bad-import-order
from . import helpers  # pylint: disable=g-bad-import-order

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------

parser = helpers.ArgumentParser()
parser.add("--batch_size", 32)
parser.add("--num_epochs", 250)
parser.add("--num_points", 2048, help="subsampled (max 2048)")
parser.add("--learning_rate", 1e-3, help="initial Adam learning rate")
parser.add("--lr_decay", True, help="enable learning rate decay")
parser.add("--bn_decay", .5, help="batch norm decay momentum")
parser.add("--tb_every", 100, help="tensorboard frequency (iterations)")
parser.add("--ev_every", 308, help="evaluation frequency (iterations)")
parser.add("--augment", True, help="use augmentations")
parser.add("--tqdm", True, help="enable the progress bar")
FLAGS = parser.parse_args()

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------

if FLAGS.lr_decay:
  lr_scheduler = tf.keras.optimizers.schedules.ExponentialDecay(
      FLAGS.learning_rate,
      decay_steps=6250,  #< 200.000 / 32 (batch size) (from original pointnet)
      decay_rate=0.7,
      staircase=True)
  optimizer = tf.keras.optimizers.Adam(learning_rate=lr_scheduler)
else:
  optimizer = tf.keras.optimizers.Adam(learning_rate=FLAGS.learning_rate)

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------

model = pointnet.PointNetVanillaClassifier(
    num_classes=40, momentum=FLAGS.bn_decay)

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------


@tf.function
def wrapped_tf_function(points, label):
  """Performs one step of minimization of the loss."""
  # --- subsampling (order DO matter)
  points = points[0:FLAGS.num_points, ...]

  # --- augmentation
  if FLAGS.augment:
    points = tf.map_fn(augment.rotate, points)
    points = augment.jitter(points)

  # --- training
  with tf.GradientTape() as tape:
    logits = model(points, training=True)
    loss = model.loss(label, logits)
  variables = model.trainable_variables
  gradients = tape.gradient(loss, variables)
  optimizer.apply_gradients(zip(gradients, variables))
  return loss


def train(example):
  """Performs one step of minimization of the loss and populates the summary."""
  points = example["points"]
  label = example["label"]
  step = optimizer.iterations.numpy()

  # --- optimize
  loss = wrapped_tf_function(points, label)
  if step % FLAGS.tb_every == 0:
    tf.summary.scalar(name="loss", data=loss, step=step)

  # --- report rate in summaries
  if FLAGS.lr_decay and step % FLAGS.tb_every == 0:
    tf.summary.scalar(name="learning_rate", data=lr_scheduler(step), step=step)


# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------


def evaluate():
  """Identify the best accuracy reached during training."""
  step = optimizer.iterations.numpy()
  if "best_accuracy" not in evaluate.__dict__:
    evaluate.best_accuracy = 0
  if step % FLAGS.ev_every != 0:
    return evaluate.best_accuracy
  aggregator = tf.keras.metrics.SparseCategoricalAccuracy()
  for example in ds_test:
    points, labels = example["points"], example["label"]
    logits = model(points, training=False)
    aggregator.update_state(labels, logits)
  accuracy = aggregator.result()
  evaluate.best_accuracy = max(accuracy, evaluate.best_accuracy)
  tf.summary.scalar(name="accuracy_test", data=accuracy, step=step)
  return evaluate.best_accuracy


# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------

ds_train, info = modelnet40.ModelNet40.load(split="train", with_info=True)
num_examples = info.splits["train"].num_examples
ds_train = ds_train.shuffle(num_examples, reshuffle_each_iteration=True)
ds_train = ds_train.repeat(FLAGS.num_epochs)
ds_train = ds_train.batch(FLAGS.batch_size)
ds_test = modelnet40.ModelNet40.load(split="test").batch(FLAGS.batch_size)

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------

try:
  helpers.setup_tensorboard(FLAGS)
  helpers.summary_command(parser, FLAGS)
  total = tf.data.experimental.cardinality(ds_train).numpy()
  pbar = tqdm.tqdm(ds_train, leave=False, total=total, disable=not FLAGS.tqdm)
  for train_example in pbar:
    train(train_example)
    best_accuracy = evaluate()
    pbar.set_postfix_str("best accuracy: {:.3f}".format(best_accuracy))

except KeyboardInterrupt:
  helpers.handle_keyboard_interrupt(FLAGS)



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Rasterization backends selector for TF Graphics."""

import enum

from tensorflow_graphics.rendering.opengl import rasterization_backend as gl_backend
from tensorflow_graphics.util import export_api


class RasterizationBackends(enum.Enum):
  OPENGL = 0


_BACKENDS = {
    RasterizationBackends.OPENGL: gl_backend,
}


def rasterize(vertices,
              triangles,
              view_projection_matrices,
              image_size,
              backend=RasterizationBackends.OPENGL):
  """Rasterizes the scene.

    This rasterizer estimates which triangle is associated with each pixel using
    OpenGL.

  Note:
    In the following, A1 to An are optional batch dimensions which must be
    broadcast compatible for inputs `vertices` and `view_projection_matrices`.

  Args:
    vertices: A tensor of shape `[A1, ..., An, V, 3]` containing batches of `V`
      vertices, each defined by a 3D point.
    triangles: A tensor of shape `[T, 3]` containing `T` triangles, each
      associated with 3 vertices from `scene_vertices`
    view_projection_matrices: A tensor of shape `[A1, ..., An, 4, 4]` containing
      batches of view projection matrices
    image_size: An tuple of integers (width, height) containing the dimensions
      in pixels of the rasterized image.
    backend: An enum containing the backend method to use for rasterization.
      Supported options are defined in the RasterizationBackends enum.

  Returns:
    A tuple of 3 elements. The first one of shape `[A1, ..., An, H, W, 1]`
    representing the triangle index associated with each pixel. If no triangle
    is associated to a pixel, the index is set to -1.
    The second element in the tuple is of shape `[A1, ..., An, H, W, 3]` and
    correspond to barycentric coordinates per pixel. The last element in the
    tuple is of shape `[A1, ..., An, H, W]` and stores a value of `0` of the
    pixel is assciated with the background, and `1` with the foreground.
  """
  return _BACKENDS[backend].rasterize(vertices, triangles,
                                      view_projection_matrices, image_size)


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements trilinear interpolation."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import shape


def interpolate(grid_3d, sampling_points, name=None):
  """Trilinear interpolation on a 3D regular grid.

  Args:
    grid_3d: A tensor with shape `[A1, ..., An, H, W, D, C]` where H, W, D are
      height, width, depth of the grid and C is the number of channels.
    sampling_points: A tensor with shape `[A1, ..., An, M, 3]` where M is the
    number of sampling points. Sampling points outside the grid are projected
    in the grid borders.
    name:  A name for this op that defaults to "trilinear_interpolate".

  Returns:
    A tensor of shape `[A1, ..., An, M, C]`
  """

  with tf.compat.v1.name_scope(name, "trilinear_interpolate",
                               [grid_3d, sampling_points]):
    grid_3d = tf.convert_to_tensor(value=grid_3d)
    sampling_points = tf.convert_to_tensor(value=sampling_points)

    shape.check_static(
        tensor=grid_3d, tensor_name="grid_3d", has_rank_greater_than=3)
    shape.check_static(tensor=sampling_points,
                       tensor_name="sampling_points",
                       has_dim_equals=(-1, 3),
                       has_rank_greater_than=1)
    shape.compare_batch_dimensions(
        tensors=(grid_3d, sampling_points),
        last_axes=(-5, -3),
        tensor_names=("grid_3d", "sampling_points"),
        broadcast_compatible=True)
    voxel_cube_shape = tf.shape(input=grid_3d)[-4:-1]
    sampling_points.set_shape(sampling_points.shape)
    batch_dims = tf.shape(input=sampling_points)[:-2]
    num_points = tf.shape(input=sampling_points)[-2]

    bottom_left = tf.floor(sampling_points)
    top_right = bottom_left + 1
    bottom_left_index = tf.cast(bottom_left, tf.int32)
    top_right_index = tf.cast(top_right, tf.int32)
    x0_index, y0_index, z0_index = tf.unstack(bottom_left_index, axis=-1)
    x1_index, y1_index, z1_index = tf.unstack(top_right_index, axis=-1)
    index_x = tf.concat([x0_index, x1_index, x0_index, x1_index,
                         x0_index, x1_index, x0_index, x1_index], axis=-1)
    index_y = tf.concat([y0_index, y0_index, y1_index, y1_index,
                         y0_index, y0_index, y1_index, y1_index], axis=-1)
    index_z = tf.concat([z0_index, z0_index, z0_index, z0_index,
                         z1_index, z1_index, z1_index, z1_index], axis=-1)
    indices = tf.stack([index_x, index_y, index_z], axis=-1)
    clip_value = tf.convert_to_tensor(value=[voxel_cube_shape - 1],
                                      dtype=indices.dtype)
    indices = tf.clip_by_value(indices, 0, clip_value)
    content = tf.gather_nd(params=grid_3d,
                           indices=indices,
                           batch_dims=tf.size(input=batch_dims))
    distance_to_bottom_left = sampling_points - bottom_left
    distance_to_top_right = top_right - sampling_points
    x_x0, y_y0, z_z0 = tf.unstack(distance_to_bottom_left, axis=-1)
    x1_x, y1_y, z1_z = tf.unstack(distance_to_top_right, axis=-1)
    weights_x = tf.concat([x1_x, x_x0, x1_x, x_x0,
                           x1_x, x_x0, x1_x, x_x0], axis=-1)
    weights_y = tf.concat([y1_y, y1_y, y_y0, y_y0,
                           y1_y, y1_y, y_y0, y_y0], axis=-1)
    weights_z = tf.concat([z1_z, z1_z, z1_z, z1_z,
                           z_z0, z_z0, z_z0, z_z0], axis=-1)
    weights = tf.expand_dims(weights_x * weights_y * weights_z, axis=-1)

    interpolated_values = weights * content
    return tf.add_n(tf.split(interpolated_values, [num_points] * 8, -2))


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Lint as: python3
"""Tests for tensorflow_graphics.datasets.features.pose_feature."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import tensorflow as tf
import tensorflow_datasets as tfds

from tensorflow_graphics.datasets.features import pose_feature


class PoseFeatureTest(tfds.testing.FeatureExpectationsTestCase):
  """Test Cases for Pose Feature Connector."""

  def test_pose_feature(self):
    expected_rotation = np.eye(3)
    expected_translation = np.zeros(3)

    expected_pose = {'R': expected_rotation.astype(np.float32),
                     't': expected_translation.astype(np.float32)}
    raising_inputs = {'rotation': expected_rotation.astype(np.float32),
                      't': expected_translation.astype(np.float32)}

    self.assertFeature(
        feature=pose_feature.Pose(),
        shape={
            'R': (3, 3),
            't': (3,)
        },
        dtype={
            'R': tf.float32,
            't': tf.float32
        },
        tests=[
            # FeaturesDict
            tfds.testing.FeatureExpectationItem(
                value=expected_pose,
                expected=expected_pose,
            ),
            tfds.testing.FeatureExpectationItem(
                value=raising_inputs,
                raise_cls=ValueError,
                raise_msg='Missing keys in provided dictionary!',
            ),
        ],
    )


if __name__ == '__main__':
  tfds.testing.test_main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the Levenberg Marquardt optimizer."""

from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.math.optimizer import levenberg_marquardt
from tensorflow_graphics.util import test_case


class OptimizerTest(test_case.TestCase):

  @parameterized.parameters(
      (4, (1,)),
      (3, (1, 1)),
      (2, (2,)),
      (1, (2, 2)),
  )
  def test_minimize_not_raised(self, max_iterations, *shapes):
    """Tests that the shape exceptions are not raised."""

    def callback(iteration, objective_value, variables):
      del iteration, objective_value, variables

    def optimize(*variables):
      levenberg_marquardt.minimize(
          lambda x: x, variables, max_iterations, callback=callback)

    self.assert_exception_is_not_raised(optimize, shapes)

  @parameterized.parameters(
      ("'max_iterations' needs to be at least 1.", -1, (1,)),
      ("'max_iterations' needs to be at least 1.", 0, (1,)),
  )
  def test_minimize_raised(self, error_msg, max_iterations, *shape):
    """Tests that the exception is raised."""

    def optimize(*variables):
      levenberg_marquardt.minimize(lambda x: x, variables, max_iterations)

    self.assert_exception_is_raised(optimize, error_msg, shape)

  @parameterized.parameters(
      ([lambda x: x], [[1.]], 1, 0., [[0.]]),
      ((lambda x: x,), ([1.],), 1, 0., [[0.]]),
      ([lambda x, y: x, lambda x, y: y], [[1.], [1.]], 1, 0., [[0.], [0.]]),
      ((lambda x, y: x, lambda x, y: y), ([1.], [1.]), 1, 0., [[0.], [0.]]),
  )
  def test_minimize_preset(self, residuals, variables, max_iterations,
                           final_objective_value, final_variables):
    """Tests the output of the minimization for some presets."""
    objective_value, output_variables = levenberg_marquardt.minimize(
        residuals, variables, max_iterations)

    with self.subTest(name="objective_value"):
      self.assertAllClose(objective_value, final_objective_value)

    with self.subTest(name="variables"):
      for output_variable, final_variable in zip(output_variables,
                                                 final_variables):
        self.assertAllClose(output_variable, final_variable)

  def test_minimize_callback(self):
    """Callback should run at the end of every iteration."""
    variables = np.random.uniform(low=-1.0, high=1.0, size=[1])
    saved_objective = [None]
    saved_iteration = [None]
    iteration_count = [0]

    def callback(iteration, objective_value, variables):
      del variables

      def save(iteration, objective_value):
        saved_objective[0] = objective_value
        saved_iteration[0] = iteration
        iteration_count[0] += 1

      return tf.py_function(save, [iteration, objective_value], [])

    final_objective, _ = levenberg_marquardt.minimize(
        lambda x: x, variables, max_iterations=5, callback=callback)
    final_saved_objective = tf.py_function(lambda: saved_objective[0], [],
                                           tf.float64)
    final_saved_iteration = tf.py_function(lambda: saved_iteration[0], [],
                                           tf.int32)
    final_iteration_count = tf.py_function(lambda: iteration_count[0], [],
                                           tf.int32)

    with self.subTest(name="objective_value"):
      self.assertAllClose(final_objective, final_saved_objective)

    with self.subTest(name="iterations"):
      self.assertAllEqual(final_saved_iteration, final_iteration_count)

  def test_minimize_ill_conditioned_not_raised(self):
    """Optimizing an ill conditioned problem should not raise an exception."""
    if not tf.executing_eagerly():
      return

    def f1(x, y):
      return x * y * 10000.0

    def f2(x, y):
      return x * y * 0.0001

    x = (1.,)
    y = (1.,)
    try:
      self.evaluate(
          levenberg_marquardt.minimize(
              residuals=(f1, f2),
              variables=(x, y),
              max_iterations=1,
              regularizer=1e-20))
    except Exception as e:  # pylint: disable=broad-except
      self.fail("Exception raised: %s" % str(e))

  def test_minimize_linear_residuals_random(self):
    """Optimizing linear residuals should give the minimum in 1 step."""
    tensor_size = np.random.randint(1, 3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    variables = np.random.uniform(low=-1.0, high=1.0, size=tensor_shape)
    objective_value, variables = levenberg_marquardt.minimize(
        lambda x: x, variables, max_iterations=1)

    with self.subTest(name="objective_value"):
      self.assertAllClose(objective_value, tf.zeros_like(objective_value))

    with self.subTest(name="variables"):
      for variable in variables:
        self.assertAllClose(variable, tf.zeros_like(variable))


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper functions for neural voxel renderer +."""
import numpy as np
import tensorflow as tf

from tensorflow_graphics.geometry.representation import grid
from tensorflow_graphics.geometry.transformation import rotation_matrix_3d
from tensorflow_graphics.math.interpolation import trilinear
from tensorflow_graphics.rendering.camera import perspective
from tensorflow_graphics.rendering.voxels import emission_absorption


X_AXIS = np.array((1., 0., 0.), dtype=np.float32)
Z_AXIS = np.array((0., 0., 1.), dtype=np.float32)

PI = np.array((-np.pi,), dtype=np.float32)
PI_2 = np.array((-np.pi/2.,), dtype=np.float32)
ZERO = np.array((0.,), dtype=np.float32)
MIRROR_X_MATRIX = np.eye(3, dtype=np.float32)
MIRROR_X_MATRIX[0, 0] = -1

# Cube that encloses all the shapes in the dataset (world coordinates)
CUBE_BOX_DIM = 1.1528184

# The bottom part of all shapes in world coordinates
OBJECT_BOTTOM = -1.1167929


def generate_ground_image(height,
                          width,
                          focal,
                          principal_point,
                          camera_rotation_matrix,
                          camera_translation_vector,
                          ground_color=(0.43, 0.43, 0.8)):
  """Generate an image depicting only the ground."""
  batch_size = camera_rotation_matrix.shape[0]
  background_image = np.ones((batch_size, height, width, 1, 1),
                             dtype=np.float32)
  background_image[:, -1, ...] = 0  # Zero the bottom line for proper sampling

  # The projection of the ground depends on the top right corner (approximation)
  plane_point_np = np.tile(np.array([[3.077984, 2.905388, 0.]],
                                    dtype=np.float32), (batch_size, 1))
  plane_point_rotated = rotation_matrix_3d.rotate(plane_point_np,
                                                  camera_rotation_matrix)
  plane_point_translated = plane_point_rotated + camera_translation_vector
  plane_point2d = \
    perspective.project(plane_point_translated, focal, principal_point)
  _, y = tf.split(plane_point2d, [1, 1], axis=-1)
  sfactor = height/y
  helper_matrix1 = np.tile(np.array([[[1, 0, 0],
                                      [0, 0, 0],
                                      [0, 0, 0]]]), (batch_size, 1, 1))
  helper_matrix2 = np.tile(np.array([[[0, 0, 0],
                                      [0, 1, 0],
                                      [0, 0, 1]]]), (batch_size, 1, 1))
  transformation_matrix = tf.multiply(tf.expand_dims(sfactor, -1),
                                      helper_matrix1) + helper_matrix2
  plane_points = grid.generate((0., 0., 0.),
                               (float(height), float(width), 0.),
                               (height, width, 1))
  plane_points = tf.reshape(plane_points, [-1, 3])
  transf_plane_points = tf.matmul(transformation_matrix,
                                  plane_points,
                                  transpose_b=True)
  interpolated_points = \
    trilinear.interpolate(background_image,
                          tf.linalg.matrix_transpose(transf_plane_points))
  ground_alpha = (1- tf.reshape(interpolated_points,
                                [batch_size, height, width, 1]))
  ground_image = tf.ones((batch_size, height, width, 3))*ground_color
  return ground_image, ground_alpha


def sampling_points_to_voxel_index(sampling_points, voxel_size):
  """Transforms the sampling points from [-1, 1] to [0, voxel_size]."""
  voxel_size = tf.convert_to_tensor(value=voxel_size)
  max_size = tf.cast(voxel_size - 1, sampling_points.dtype)
  return 0.5 * ((sampling_points + 1) * max_size)


def sampling_points_from_3d_grid(grid_size, dtype=tf.float32):
  """Returns a tensor of shape `[M, 3]`, with M the number of sampling points."""
  sampling_points = grid.generate((-1.0, -1.0, -1.0), (1.0, 1.0, 1.0),
                                  grid_size)
  sampling_points = tf.cast(sampling_points, dtype)
  return tf.reshape(sampling_points, [-1, 3])


def sampling_points_from_frustum(height,
                                 width,
                                 focal,
                                 principal_point,
                                 depth_min=0.0,
                                 depth_max=5.,
                                 frustum_size=(256, 256, 256)):
  """Generates samples from a camera frustum."""

  # ------------------ Get the rays from the camera ----------------------------
  sampling_points = grid.generate((0., 0.),
                                  (float(width)-1, float(height)-1),
                                  (frustum_size[0], frustum_size[1]))
  sampling_points = tf.reshape(sampling_points, [-1, 2])  # [h*w, 2]
  rays = perspective.ray(sampling_points, focal, principal_point)  # [h*w, 3]

  # ------------------ Extract a volume in front of the camera -----------------
  depth_tensor = grid.generate((depth_min,), (depth_max,), (frustum_size[2],))
  sampling_volume = tf.multiply(tf.expand_dims(rays, axis=-1),
                                tf.transpose(depth_tensor))  # [h*w, 3, dstep]
  sampling_volume = tf.transpose(sampling_volume, [0, 2, 1])  # [h*w, dstep, 3]
  sampling_volume = tf.reshape(sampling_volume, [-1, 3])  # [h*w*dstep, 3]
  return sampling_volume


def place_frustum_sampling_points_at_blender_camera(sampling_points,
                                                    camera_rotation_matrix,
                                                    camera_translation_vector):
  """Transform the TF-Graphics frustum points to the blender camera position."""
  blender_rotation_matrix = rotation_matrix_3d.from_axis_angle(X_AXIS, -PI)
  camera_translation_vector = tf.matmul(blender_rotation_matrix,
                                        camera_translation_vector)

  # The original frustum points are mirrored in the x axis and rotated 270
  mirror_x_matrix = tf.convert_to_tensor(MIRROR_X_MATRIX)
  up_rotation_matrix = rotation_matrix_3d.from_axis_angle(Z_AXIS, 3*PI_2)
  local_camera_transform = tf.matmul(mirror_x_matrix, up_rotation_matrix)

  sampling_points = tf.matmul(local_camera_transform,
                              sampling_points,
                              transpose_b=True)
  sampling_points = \
    tf.matmul(tf.linalg.matrix_transpose(camera_rotation_matrix),
              sampling_points+camera_translation_vector)
  return tf.linalg.matrix_transpose(sampling_points)  # [h*w*dstep, 3]


def transform_volume(voxels, transformation_matrix, voxel_size=(128, 128, 128)):
  """Apply a transformation to the input voxels."""
  voxels = tf.convert_to_tensor(voxels)
  volume_sampling = sampling_points_from_3d_grid(voxel_size)
  volume_sampling = tf.matmul(transformation_matrix,
                              tf.transpose(a=volume_sampling))
  volume_sampling = tf.cast(tf.linalg.matrix_transpose(volume_sampling),
                            tf.float32)
  volume_sampling = sampling_points_to_voxel_index(volume_sampling, voxel_size)
  interpolated_voxels = trilinear.interpolate(voxels, volume_sampling)
  return tf.reshape(interpolated_voxels, tf.shape(voxels))


def object_rotation_in_blender_world(voxels, object_rotation):
  """Rotate the voxels as in blender world."""
  euler_angles = np.array([0, 0, 1], dtype=np.float32)*np.deg2rad(90)
  object_correction_matrix = rotation_matrix_3d.from_euler(euler_angles)
  euler_angles = np.array([0, 1, 0], dtype=np.float32)*(-object_rotation)
  object_rotation_matrix = rotation_matrix_3d.from_euler(euler_angles)
  euler_angles_blender = np.array([1, 0, 0], dtype=np.float32)*np.deg2rad(-90)
  blender_object_correction_matrix = \
    rotation_matrix_3d.from_euler(euler_angles_blender)
  transformation_matrix = tf.matmul(tf.matmul(object_correction_matrix,
                                              object_rotation_matrix),
                                    blender_object_correction_matrix)

  return transform_volume(voxels, transformation_matrix)


def render_voxels_from_blender_camera(voxels,
                                      object_rotation,
                                      object_translation,
                                      height,
                                      width,
                                      focal,
                                      principal_point,
                                      camera_rotation_matrix,
                                      camera_translation_vector,
                                      frustum_size=(256, 256, 512),
                                      absorption_factor=0.1,
                                      cell_size=1.0,
                                      depth_min=0.0,
                                      depth_max=5.0):
  """Renders the voxels according to their position in the world."""
  batch_size = voxels.shape[0]
  voxel_size = voxels.shape[1]
  sampling_volume = sampling_points_from_frustum(height,
                                                 width,
                                                 focal,
                                                 principal_point,
                                                 depth_min=depth_min,
                                                 depth_max=depth_max,
                                                 frustum_size=frustum_size)
  sampling_volume = \
    place_frustum_sampling_points_at_blender_camera(sampling_volume,
                                                    camera_rotation_matrix,
                                                    camera_translation_vector)

  interpolated_voxels = \
    object_rotation_in_blender_world(voxels, object_rotation)

  # Adjust the camera (translate the camera instead of the object)
  sampling_volume = sampling_volume - object_translation
  sampling_volume = sampling_volume/CUBE_BOX_DIM
  sampling_volume = sampling_points_to_voxel_index(sampling_volume, voxel_size)

  camera_voxels = trilinear.interpolate(interpolated_voxels, sampling_volume)
  camera_voxels = tf.reshape(camera_voxels,
                             [batch_size] + list(frustum_size) + [4])
  voxel_image = emission_absorption.render(camera_voxels,
                                           absorption_factor=absorption_factor,
                                           cell_size=cell_size)
  return voxel_image


def object_to_world(voxels,
                    euler_angles_x,
                    euler_angles_y,
                    translation_vector,
                    target_volume_size=(128, 128, 128)):
  """Apply the transformations to the voxels and place them in world coords."""
  scale_factor = 1.82  # object to world voxel space scale factor

  translation_vector = tf.expand_dims(translation_vector, axis=-1)

  sampling_points = tf.cast(sampling_points_from_3d_grid(target_volume_size),
                            tf.float32)  # 128^3 X 3
  transf_matrix_x = rotation_matrix_3d.from_euler(euler_angles_x)  # [B, 3, 3]
  transf_matrix_y = rotation_matrix_3d.from_euler(euler_angles_y)  # [B, 3, 3]
  transf_matrix = tf.matmul(transf_matrix_x, transf_matrix_y)  # [B, 3, 3]
  transf_matrix = transf_matrix*scale_factor  # [B, 3, 3]
  sampling_points = tf.matmul(transf_matrix,
                              tf.transpose(sampling_points))  # [B, 3, N]
  translation_vector = tf.matmul(transf_matrix, translation_vector)  # [B, 3, 1]
  sampling_points = sampling_points - translation_vector
  sampling_points = tf.linalg.matrix_transpose(sampling_points)
  sampling_points = sampling_points_to_voxel_index(sampling_points,
                                                   target_volume_size)
  sampling_points = tf.cast(sampling_points, tf.float32)
  interpolated_points = trilinear.interpolate(voxels, sampling_points)
  interpolated_voxels = tf.reshape(interpolated_points,
                                   [-1] + list(target_volume_size)+[4])

  return interpolated_voxels



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements the chamfer distance."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import shape


def evaluate(point_set_a, point_set_b, name=None):
  """Computes the Chamfer distance for the given two point sets.

  Note:
    This is a symmetric version of the Chamfer distance, calculated as the sum
    of the average minimum distance from point_set_a to point_set_b and vice
    versa.
    The average minimum distance from one point set to another is calculated as
    the average of the distances between the points in the first set and their
    closest point in the second set, and is thus not symmetrical.

  Note:
    This function returns the exact Chamfer distance and not an approximation.

  Note:
    In the following, A1 to An are optional batch dimensions, which must be
    broadcast compatible.

  Args:
    point_set_a: A tensor of shape `[A1, ..., An, N, D]`, where the last axis
      represents points in a D dimensional space.
    point_set_b: A tensor of shape `[A1, ..., An, M, D]`, where the last axis
      represents points in a D dimensional space.
    name: A name for this op. Defaults to "chamfer_distance_evaluate".

  Returns:
    A tensor of shape `[A1, ..., An]` storing the chamfer distance between the
    two point sets.

  Raises:
    ValueError: if the shape of `point_set_a`, `point_set_b` is not supported.
  """
  with tf.compat.v1.name_scope(name, "chamfer_distance_evaluate",
                               [point_set_a, point_set_b]):
    point_set_a = tf.convert_to_tensor(value=point_set_a)
    point_set_b = tf.convert_to_tensor(value=point_set_b)

    shape.compare_batch_dimensions(
        tensors=(point_set_a, point_set_b),
        tensor_names=("point_set_a", "point_set_b"),
        last_axes=-3,
        broadcast_compatible=True)
    # Verify that the last axis of the tensors has the same dimension.
    dimension = point_set_a.shape.as_list()[-1]
    shape.check_static(
        tensor=point_set_b,
        tensor_name="point_set_b",
        has_dim_equals=(-1, dimension))

    # Create N x M matrix where the entry i,j corresponds to ai - bj (vector of
    # dimension D).
    difference = (
        tf.expand_dims(point_set_a, axis=-2) -
        tf.expand_dims(point_set_b, axis=-3))
    # Calculate the square distances between each two points: |ai - bj|^2.
    square_distances = tf.einsum("...i,...i->...", difference, difference)

    minimum_square_distance_a_to_b = tf.reduce_min(
        input_tensor=square_distances, axis=-1)
    minimum_square_distance_b_to_a = tf.reduce_min(
        input_tensor=square_distances, axis=-2)

    return (
        tf.reduce_mean(input_tensor=minimum_square_distance_a_to_b, axis=-1) +
        tf.reduce_mean(input_tensor=minimum_square_distance_b_to_a, axis=-1))


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for axis-angle."""

from absl.testing import flagsaver
from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.geometry.transformation import axis_angle
from tensorflow_graphics.geometry.transformation import quaternion
from tensorflow_graphics.geometry.transformation import rotation_matrix_3d
from tensorflow_graphics.geometry.transformation.tests import test_helpers
from tensorflow_graphics.util import test_case


class AxisAngleTest(test_case.TestCase):

  @parameterized.parameters(
      ((3,),),
      ((None, 3),),
  )
  def test_from_euler_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(axis_angle.from_euler, shapes)

  @parameterized.parameters(
      ("must have exactly 3 dimensions in axis -1", (None,)),)
  def test_from_euler_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(axis_angle.from_euler, error_msg, shapes)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_from_euler_jacobian_random(self):
    """Test the Jacobian of the from_euler function.

    Note:
      Preset angles are not tested as the gradient of tf.norm is NaN at 0.
    """
    x_init = test_helpers.generate_random_test_euler_angles()

    self.assert_jacobian_is_finite_fn(lambda x: axis_angle.from_euler(x)[0],
                                      [x_init])
    self.assert_jacobian_is_finite_fn(lambda x: axis_angle.from_euler(x)[1],
                                      [x_init])

  def test_from_euler_random(self):
    """Tests that from_euler allows to perform the expect rotation of points."""
    random_euler_angles = test_helpers.generate_random_test_euler_angles()
    tensor_shape = random_euler_angles.shape[:-1]
    random_point = np.random.normal(size=tensor_shape + (3,))

    random_matrix = rotation_matrix_3d.from_euler(random_euler_angles)
    random_axis, random_angle = axis_angle.from_euler(random_euler_angles)
    rotated_with_matrix = rotation_matrix_3d.rotate(random_point, random_matrix)
    rotated_with_axis_angle = axis_angle.rotate(random_point, random_axis,
                                                random_angle)

    self.assertAllClose(rotated_with_matrix, rotated_with_axis_angle)

  @parameterized.parameters(
      ((3,),),
      ((None, 3),),
      ((2, 3),),
  )
  def test_from_euler_with_small_angles_approximation_exception_not_raised(
      self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(
        axis_angle.from_euler_with_small_angles_approximation, shapes)

  @parameterized.parameters(
      ("must have exactly 3 dimensions in axis -1", (None,)),)
  def test_from_euler_with_small_angles_approximation_exception_raised(
      self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(
        axis_angle.from_euler_with_small_angles_approximation, error_msg,
        shapes)

  def test_from_euler_normalized_preset(self):
    """Tests that from_euler allows build normalized axis-angles."""
    euler_angles = test_helpers.generate_preset_test_euler_angles()

    axis, angle = axis_angle.from_euler(euler_angles)

    self.assertAllEqual(
        axis_angle.is_normalized(axis, angle), np.ones(angle.shape, dtype=bool))

  def test_from_euler_normalized_random(self):
    """Tests that from_euler allows build normalized axis-angles."""
    random_euler_angles = test_helpers.generate_random_test_euler_angles()

    random_axis, random_angle = axis_angle.from_euler(random_euler_angles)

    self.assertAllEqual(
        axis_angle.is_normalized(random_axis, random_angle),
        np.ones(shape=random_angle.shape))

  def test_from_euler_with_small_angles_approximation_random(self):
    # Only generate small angles. For a test tolerance of 1e-3, 0.23 was found
    # empirically to be the range where the small angle approximation works.
    random_euler_angles = test_helpers.generate_random_test_euler_angles(
        min_angle=-0.23, max_angle=0.23)

    exact_axis_angle = axis_angle.from_euler(random_euler_angles)
    approximate_axis_angle = (
        axis_angle.from_euler_with_small_angles_approximation(
            random_euler_angles))

    self.assertAllClose(exact_axis_angle, approximate_axis_angle, atol=1e-3)

  @parameterized.parameters(
      ((4,),),
      ((None, 4),),
      ((2, 4),),
  )
  def test_from_quaternion_exception_not_raised(self, *shape):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(axis_angle.from_quaternion, shape)

  @parameterized.parameters(
      ("must have exactly 4 dimensions in axis -1", (None,)),)
  def test_from_quaternion_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(axis_angle.from_quaternion, error_msg,
                                    shape)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_from_quaternion_jacobian_random(self):
    """Test the Jacobian of the from_quaternion function.

    Note:
      Preset angles are not tested as the gradient of tf.norm is NaN a 0.
    """
    x_init = test_helpers.generate_random_test_quaternions()

    self.assert_jacobian_is_finite_fn(
        lambda x: axis_angle.from_quaternion(x)[0], [x_init])
    self.assert_jacobian_is_finite_fn(
        lambda x: axis_angle.from_quaternion(x)[1], [x_init])

  def test_from_quaternion_normalized_preset(self):
    """Tests that from_quaternion returns normalized axis-angles."""
    euler_angles = test_helpers.generate_preset_test_euler_angles()

    quat = quaternion.from_euler(euler_angles)
    axis, angle = axis_angle.from_quaternion(quat)

    self.assertAllEqual(
        axis_angle.is_normalized(axis, angle), np.ones(angle.shape, dtype=bool))

  def test_from_quaternion_normalized_random(self):
    """Tests that from_quaternion returns normalized axis-angles."""
    random_quaternions = test_helpers.generate_random_test_quaternions()

    random_axis, random_angle = axis_angle.from_quaternion(random_quaternions)

    self.assertAllEqual(
        axis_angle.is_normalized(random_axis, random_angle),
        np.ones(random_angle.shape))

  def test_from_quaternion_preset(self):
    """Tests that axis_angle.from_quaternion produces the expected result."""
    preset_euler_angles = test_helpers.generate_preset_test_euler_angles()

    preset_quaternions = quaternion.from_euler(preset_euler_angles)
    preset_axis_angle = axis_angle.from_euler(preset_euler_angles)

    self.assertAllClose(
        preset_axis_angle,
        axis_angle.from_quaternion(preset_quaternions),
        rtol=1e-3)

  def test_from_quaternion_random(self):
    """Tests that axis_angle.from_quaternion produces the expected result."""
    random_euler_angles = test_helpers.generate_random_test_euler_angles()

    random_quaternions = quaternion.from_euler(random_euler_angles)
    random_axis_angle = axis_angle.from_euler(random_euler_angles)

    self.assertAllClose(
        random_axis_angle,
        axis_angle.from_quaternion(random_quaternions),
        rtol=1e-3)

  @parameterized.parameters(
      ((3, 3),),
      ((None, 3, 3),),
  )
  def test_from_rotation_matrix_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(axis_angle.from_rotation_matrix, shapes)

  @parameterized.parameters(
      ("must have a rank greater than 1", (3,)),
      ("must have exactly 3 dimensions in axis -1", (3, None)),
      ("must have exactly 3 dimensions in axis -2", (None, 3)),
  )
  def test_from_rotation_matrix_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(axis_angle.from_rotation_matrix, error_msg,
                                    shape)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_from_rotation_matrix_jacobian_random(self):
    """Test the Jacobian of the from_rotation_matrix function.

    Note:
      Preset angles are not tested as the gradient of tf.norm is NaN a 0.
    """
    x_init = test_helpers.generate_random_test_rotation_matrix_3d()

    self.assert_jacobian_is_finite_fn(
        lambda x: axis_angle.from_rotation_matrix(x)[0], [x_init])
    self.assert_jacobian_is_finite_fn(
        lambda x: axis_angle.from_rotation_matrix(x)[1], [x_init])

  def test_from_rotation_matrix_normalized_preset(self):
    """Tests that from_rotation_matrix returns normalized axis-angles."""
    preset_euler_angles = test_helpers.generate_preset_test_euler_angles()

    matrix = rotation_matrix_3d.from_euler(preset_euler_angles)
    axis, angle = axis_angle.from_rotation_matrix(matrix)

    self.assertAllEqual(
        axis_angle.is_normalized(axis, angle), np.ones(angle.shape, dtype=bool))

  def test_from_rotation_matrix_normalized_random(self):
    """Tests that from_rotation_matrix returns normalized axis-angles."""
    random_euler_angles = test_helpers.generate_random_test_euler_angles()

    matrix = rotation_matrix_3d.from_euler(random_euler_angles)
    axis, angle = axis_angle.from_rotation_matrix(matrix)

    self.assertAllEqual(
        axis_angle.is_normalized(axis, angle), np.ones(angle.shape, dtype=bool))

  def test_from_rotation_matrix_random(self):
    """Tests rotation around Z axis."""

    def get_rotation_matrix_around_z(angle_rad):
      return np.array([
          [np.cos(angle_rad), -np.sin(angle_rad), 0],
          [np.sin(angle_rad), np.cos(angle_rad), 0],
          [0, 0, 1],
      ])

    tensor_size = np.random.randint(10)
    angle = (
        np.array([
            np.deg2rad(np.random.randint(720) - 360) for _ in range(tensor_size)
        ]).reshape((tensor_size, 1)))
    rotation_matrix = [get_rotation_matrix_around_z(i[0]) for i in angle]
    rotation_matrix = np.array(rotation_matrix).reshape((tensor_size, 3, 3))
    tf_axis, tf_angle = axis_angle.from_rotation_matrix(rotation_matrix)
    axis = np.tile([[0., 0., 1.]], (angle.shape[0], 1))
    tf_quat_gt = quaternion.from_axis_angle(axis, angle)
    tf_quat = quaternion.from_axis_angle(tf_axis, tf_angle)
    # Compare quaternions since axis orientation and angle ambiguity will
    # lead to more complex comparisons.
    for quat_gt, quat in zip(self.evaluate(tf_quat_gt), self.evaluate(tf_quat)):
      # Remember that q=-q for any quaternion.
      pos = np.allclose(quat_gt, quat)
      neg = np.allclose(quat_gt, -quat)

      self.assertTrue(pos or neg)

  @parameterized.parameters(
      ((3,), (1,)),
      ((None, 3), (None, 1)),
      ((2, 3), (2, 1)),
      ((1, 3), (1,)),
      ((3,), (1, 1)),
  )
  def test_inverse_exception_not_raised(self, *shape):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(axis_angle.inverse, shape)

  @parameterized.parameters(
      ("must have exactly 3 dimensions in axis -1", (None,), (1,)),
      ("must have exactly 1 dimensions in axis -1", (3,), (None,)),
  )
  def test_inverse_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(axis_angle.inverse, error_msg, shape)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_inverse_jacobian_preset(self):
    """Test the Jacobian of the inverse function."""
    x_axis_init, x_angle_init = test_helpers.generate_preset_test_axis_angle()

    if tf.executing_eagerly():
      # Because axis is returned as is, gradient calculation fails in graph mode
      # but not in eager mode. This is a side effect of having a graph rather
      # than a problem of the function.
      with self.subTest("axis"):
        self.assert_jacobian_is_correct_fn(
            lambda x: axis_angle.inverse(x, x_angle_init)[0], [x_axis_init])

    with self.subTest("angle"):
      self.assert_jacobian_is_correct_fn(
          lambda x: axis_angle.inverse(x_axis_init, x)[1], [x_angle_init])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_inverse_jacobian_random(self):
    """Test the Jacobian of the inverse function."""
    x_axis_init, x_angle_init = test_helpers.generate_random_test_axis_angle()

    if tf.executing_eagerly():
      # Because axis is returned as is, gradient calculation fails in graph mode
      # but not in eager mode. This is a side effect of having a graph rather
      # than a problem of the function.
      with self.subTest("axis"):
        self.assert_jacobian_is_correct_fn(
            lambda x: axis_angle.inverse(1.0 * x, x_angle_init)[0],
            [x_axis_init])

    with self.subTest("angle"):
      self.assert_jacobian_is_correct_fn(
          lambda x: axis_angle.inverse(x_axis_init, x)[1], [x_angle_init])

  def test_inverse_normalized_random(self):
    """Tests that axis-angle inversion return a normalized axis-angle."""
    random_axis, random_angle = test_helpers.generate_random_test_axis_angle()

    inverse_axis, inverse_angle = axis_angle.inverse(random_axis, random_angle)

    self.assertAllEqual(
        axis_angle.is_normalized(inverse_axis, inverse_angle),
        np.ones(random_angle.shape))

  def test_inverse_random(self):
    """Tests axis-angle inversion."""
    random_axis, random_angle = test_helpers.generate_random_test_axis_angle()

    inverse_axis, inverse_angle = axis_angle.inverse(random_axis, random_angle)

    self.assertAllClose(inverse_axis, random_axis, rtol=1e-3)
    self.assertAllClose(inverse_angle, -random_angle, rtol=1e-3)

  @parameterized.parameters(
      ("must have exactly 3 dimensions in axis -1", (None,), (1,)),
      ("must have exactly 1 dimensions in axis -1", (3,), (None,)),
  )
  def test_is_normalized_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(axis_angle.is_normalized, error_msg, shape)

  def test_is_normalized_random(self):
    """Tests that is_normalized works as intended."""
    # Samples normalized axis-angles.
    random_euler_angles = test_helpers.generate_random_test_euler_angles()

    with self.subTest(name=("is_normalized")):
      random_axis, random_angle = axis_angle.from_euler(random_euler_angles)
      pred = axis_angle.is_normalized(random_axis, random_angle)

      self.assertAllEqual(np.ones(shape=random_angle.shape, dtype=bool), pred)

    with self.subTest(name=("is_not_normalized")):
      random_axis *= 1.01
      pred = axis_angle.is_normalized(random_axis, random_angle)

      self.assertAllEqual(np.zeros(shape=random_angle.shape, dtype=bool), pred)

  @parameterized.parameters(
      ((3,), (3,), (1,)),
      ((None, 3), (None, 3), (None, 1)),
      ((2, 3), (2, 3), (2, 1)),
      ((3,), (1, 3), (1, 2, 1)),
      ((1, 2, 3), (1, 3), (1,)),
      ((3,), (1, 3), (1,)),
  )
  def test_rotate_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(axis_angle.rotate, shapes)

  @parameterized.parameters(
      ("must have exactly 3 dimensions in axis -1", (2,), (3,), (1,)),
      ("must have exactly 3 dimensions in axis -1", (3,), (2,), (1,)),
      ("must have exactly 1 dimensions in axis -1", (3,), (3,), (2,)),
  )
  def test_rotate_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(axis_angle.rotate, error_msg, shape)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_rotate_jacobian_preset(self):
    """Test the Jacobian of the rotate function."""
    x_axis_init, x_angle_init = test_helpers.generate_preset_test_axis_angle()
    x_point_init = np.random.uniform(size=x_axis_init.shape)

    self.assert_jacobian_is_correct_fn(
        axis_angle.rotate, [x_point_init, x_axis_init, x_angle_init])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_rotate_jacobian_random(self):
    """Test the Jacobian of the rotate function."""
    x_axis_init, x_angle_init = test_helpers.generate_random_test_axis_angle()
    x_point_init = np.random.uniform(size=x_axis_init.shape)

    self.assert_jacobian_is_correct_fn(
        axis_angle.rotate, [x_point_init, x_axis_init, x_angle_init])

  def test_rotate_random(self):
    """Tests that the rotate provide the same results as quaternion.rotate."""
    random_axis, random_angle = test_helpers.generate_random_test_axis_angle()
    tensor_shape = random_angle.shape[:-1]
    random_point = np.random.normal(size=tensor_shape + (3,))

    random_quaternion = quaternion.from_axis_angle(random_axis, random_angle)
    ground_truth = quaternion.rotate(random_point, random_quaternion)
    prediction = axis_angle.rotate(random_point, random_axis, random_angle)

    self.assertAllClose(ground_truth, prediction, rtol=1e-6)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements weighted interpolation for point sets."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow_graphics.math import vector
from tensorflow_graphics.util import asserts
from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import safe_ops
from tensorflow_graphics.util import shape


def interpolate(points,
                weights,
                indices,
                normalize=True,
                allow_negative_weights=False,
                name=None):
  """Weighted interpolation for M-D point sets.

  Given an M-D point set, this function can be used to generate a new point set
  that is formed by interpolating a subset of points in the set.

  Note:
    In the following, A1 to An, and B1 to Bk are optional batch dimensions.

  Args:
    points: A tensor with shape `[B1, ..., Bk, M] and rank R > 1, where M is the
      dimensionality of the points.
    weights: A tensor with shape `[A1, ..., An, P]`, where P is the number of
      points to interpolate for each output point.
    indices: A tensor of dtype tf.int32 and shape `[A1, ..., An, P, R-1]`, which
      contains the point indices to be used for each output point. The R-1
      dimensional axis gives the slice index of a single point in `points`. The
      first n+1 dimensions of weights and indices must match, or be broadcast
      compatible.
    normalize: A `bool` describing whether or not to normalize the weights on
      the last axis.
    allow_negative_weights: A `bool` describing whether or not negative weights
      are allowed.
    name: A name for this op. Defaults to "weighted_interpolate".

  Returns:
    A tensor of shape `[A1, ..., An, M]` storing the interpolated M-D
    points. The first n dimensions will be the same as weights and indices.
  """
  with tf.compat.v1.name_scope(name, "weighted_interpolate",
                               [points, weights, indices]):
    points = tf.convert_to_tensor(value=points)
    weights = tf.convert_to_tensor(value=weights)
    indices = tf.convert_to_tensor(value=indices)

    shape.check_static(
        tensor=points, tensor_name="points", has_rank_greater_than=1)
    shape.check_static(
        tensor=indices,
        tensor_name="indices",
        has_rank_greater_than=1,
        has_dim_equals=(-1, points.shape.ndims - 1))
    shape.compare_dimensions(
        tensors=(weights, indices),
        axes=(-1, -2),
        tensor_names=("weights", "indices"))
    shape.compare_batch_dimensions(
        tensors=(weights, indices),
        last_axes=(-2, -3),
        tensor_names=("weights", "indices"),
        broadcast_compatible=True)
    if not allow_negative_weights:
      weights = asserts.assert_all_above(weights, 0.0, open_bound=False)

    if normalize:
      sums = tf.reduce_sum(input_tensor=weights, axis=-1, keepdims=True)
      sums = asserts.assert_nonzero_norm(sums)
      weights = safe_ops.safe_signed_div(weights, sums)
    point_lists = tf.gather_nd(points, indices)
    return vector.dot(
        point_lists, tf.expand_dims(weights, axis=-1), axis=-2, keepdims=False)


def get_barycentric_coordinates(triangle_vertices, pixels, name=None):
  """Computes the barycentric coordinates of pixels for 2D triangles.

  Barycentric coordinates of a point `p` are represented as coefficients
  $(w_1, w_2, w_3)$ corresponding to the masses placed at the vertices of a
  reference triangle if `p` is the center of mass. Barycentric coordinates are
  normalized so that $w_1 + w_2 + w_3 = 1$. These coordinates play an essential
  role in computing the pixel attributes (e.g. depth, color, normals, and
  texture coordinates) of a point lying on the surface of a triangle. The point
  `p` is inside the triangle if all of its barycentric coordinates are positive.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    triangle_vertices: A tensor of shape `[A1, ..., An, 3, 2]`, where the last
      two dimensions represents the `x` and `y` coordinates for each vertex of a
      2D triangle.
    pixels: A tensor of shape `[A1, ..., An, N, 2]`, where `N` represents the
      number of pixels, and the last dimension represents the `x` and `y`
      coordinates of each pixel.
    name: A name for this op that defaults to
      "rasterizer_get_barycentric_coordinates".

  Returns:
    barycentric_coordinates: A float tensor of shape `[A1, ..., An, N, 3]`,
      representing the barycentric coordinates.
    valid: A boolean tensor of shape `[A1, ..., An, N], which is `True` where
      pixels are inside the triangle, and `False` otherwise.
  """
  with tf.compat.v1.name_scope(name, "rasterizer_get_barycentric_coordinates",
                               [triangle_vertices, pixels]):
    triangle_vertices = tf.convert_to_tensor(value=triangle_vertices)
    pixels = tf.convert_to_tensor(value=pixels)

    shape.check_static(
        tensor=triangle_vertices,
        tensor_name="triangle_vertices",
        has_dim_equals=((-1, 2), (-2, 3)))
    shape.check_static(
        tensor=pixels, tensor_name="pixels", has_dim_equals=(-1, 2))
    shape.compare_batch_dimensions(
        tensors=(triangle_vertices, pixels),
        last_axes=(-3, -3),
        broadcast_compatible=True)

    vertex_1, vertex_2, vertex_3 = tf.unstack(
        tf.expand_dims(triangle_vertices, axis=-3), axis=-2)
    vertex_x1, vertex_y1 = tf.unstack(vertex_1, axis=-1)
    vertex_x2, vertex_y2 = tf.unstack(vertex_2, axis=-1)
    vertex_x3, vertex_y3 = tf.unstack(vertex_3, axis=-1)
    pixels_x, pixels_y = tf.unstack(pixels, axis=-1)

    x1_minus_x3 = vertex_x1 - vertex_x3
    x3_minus_x2 = vertex_x3 - vertex_x2
    y3_minus_y1 = vertex_y3 - vertex_y1
    y2_minus_y3 = vertex_y2 - vertex_y3
    x_minus_x3 = pixels_x - vertex_x3
    y_minus_y3 = pixels_y - vertex_y3

    determinant = y2_minus_y3 * x1_minus_x3 - x3_minus_x2 * y3_minus_y1
    coordinate_1 = y2_minus_y3 * x_minus_x3 + x3_minus_x2 * y_minus_y3
    coordinate_1 = safe_ops.safe_signed_div(coordinate_1, determinant)
    coordinate_2 = y3_minus_y1 * x_minus_x3 + x1_minus_x3 * y_minus_y3
    coordinate_2 = safe_ops.safe_signed_div(coordinate_2, determinant)
    coordinate_3 = 1.0 - (coordinate_1 + coordinate_2)

    barycentric_coordinates = tf.stack(
        (coordinate_1, coordinate_2, coordinate_3), axis=-1)
    valid = tf.logical_and(
        tf.logical_and(coordinate_1 >= 0.0, coordinate_2 >= 0.0),
        coordinate_3 >= 0.0)
    return barycentric_coordinates, valid

# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for tensorflow_graphics.geometry.convolution.tests.graph_pooling."""

# pylint: disable=protected-access

import itertools
from absl.testing import parameterized
import numpy as np
import tensorflow as tf

import tensorflow_graphics.geometry.convolution.graph_pooling as gp
from tensorflow_graphics.geometry.convolution.tests import utils_test
from tensorflow_graphics.util import test_case


def _dense_to_sparse(data):
  """Convert a numpy array to a tf.SparseTensor."""
  return utils_test._dense_to_sparse(data)


def _batch_sparse_eye(batch_shape, num_vertices, dtype):
  """Generate a batch of identity matrices."""
  eye = np.eye(num_vertices, dtype=dtype)
  num_batch_dims = len(batch_shape)
  expand_shape = np.concatenate((np.ones(
      num_batch_dims, dtype=np.int32), (num_vertices, num_vertices)),
                                axis=0)
  eye = np.reshape(eye, expand_shape)
  tile_shape = np.concatenate((batch_shape, (1, 1)), axis=0)
  return _dense_to_sparse(np.tile(eye, tile_shape))


class GraphPoolingTestPoolTests(test_case.TestCase):

  @parameterized.parameters(
      ("'sizes' must have an integer type.", np.float32, np.float32,
       np.float32),
      ("'data' must have a float type.", np.int32, np.float32, np.int32),
      ("'pool_map' and 'data' must have the same type.", np.float32, np.float64,
       np.int32))
  def test_pool_exception_raised_types(self, err_msg, data_type, pool_map_type,
                                       sizes_type):
    """Tests the correct exceptions are raised for invalid types."""
    data = np.ones((2, 3, 3), dtype=data_type)
    pool_map = _dense_to_sparse(np.ones((2, 3, 3), dtype=pool_map_type))
    sizes = np.array(((1, 2), (2, 3)), dtype=sizes_type)

    with self.assertRaisesRegexp(TypeError, err_msg):
      gp.pool(data, pool_map, sizes)

  @parameterized.parameters(
      ('data must have a rank greater than 1', (3,), (3,), None),
      ('pool_map must have a rank of 2', (3, 3), (3,), None),
      ('sizes must have a rank of 3', (4, 5, 3, 2), (4, 5, 3, 3), (3, 2)),
  )
  def test_pool_exception_raised_shapes(self, err_msg, data_shape,
                                        pool_map_shape, sizes_shape):
    """Tests the correct exceptions are raised for invalid shapes."""
    data = np.ones(data_shape, dtype=np.float32)
    pool_map = _dense_to_sparse(np.ones(pool_map_shape, dtype=np.float32))
    if sizes_shape is not None:
      sizes = np.ones(sizes_shape, dtype=np.int32)
    else:
      sizes = None

    with self.assertRaisesRegexp(ValueError, err_msg):
      gp.pool(data, pool_map, sizes)

  def test_pool_exception_raised_algorithm(self):
    """Tests the correct exception is raised for an invalid algorithm."""
    data = np.ones(shape=(2, 2))
    pool_map = _dense_to_sparse(np.ones(shape=(2, 2)))

    with self.assertRaisesRegexp(
        ValueError, 'The pooling method must be "weighted" or "max"'):
      gp.pool(data, pool_map, sizes=None, algorithm='mean')

  @parameterized.parameters(
      ((2, 3), 4, 3, np.float32),
      ((1,), 6, 1, np.float32),
      ((4, 1, 3), 9, 7, np.float64),
      ((2, 8, 4, 6), 19, 11, np.float64),
  )
  def test_pool_identity(self, batch_shape, num_vertices, num_features,
                         data_type):
    """Tests graph pooling with identity maps."""
    data_shape = np.concatenate((batch_shape, (num_vertices, num_features)))
    data = np.random.uniform(size=data_shape).astype(data_type)
    pool_map = _batch_sparse_eye(batch_shape, num_vertices, data_type)

    pooled_max = gp.pool(data, pool_map, sizes=None, algorithm='max', name=None)
    pooled_weighted = gp.pool(
        data, pool_map, sizes=None, algorithm='weighted', name=None)

    self.assertAllClose(pooled_max, data)
    self.assertAllClose(pooled_weighted, data)

  def test_pool_preset_padded(self):
    """Tests pooling with preset data and padding."""
    data = np.reshape(np.arange(12).astype(np.float32), (2, 3, 2))
    sizes = ((2, 3), (3, 3))
    pool_map = _dense_to_sparse(
        np.array((((0.5, 0.5, 0.), (0., 0., 1.), (0., 0., 0.)),
                  ((1., 0., 0.), (0., 1., 0.), (0., 0., 1.))),
                 dtype=np.float32))

    pooled_max = gp.pool(data, pool_map, sizes, algorithm='max')
    pooled_weighted = gp.pool(data, pool_map, sizes, algorithm='weighted')
    true_max = (((2., 3.), (4., 5.), (0., 0.)), ((6., 7.), (8., 9.), (10.,
                                                                      11.)))
    true_weighted = (((1., 2.), (4., 5.), (0., 0.)), ((6., 7.), (8., 9.),
                                                      (10., 11.)))

    self.assertAllClose(pooled_max, true_max)
    self.assertAllClose(pooled_weighted, true_weighted)

  def test_pool_preset(self):
    """Tests pooling with preset data."""
    pool_map = np.array(((0.5, 0.5, 0., 0.), (0., 0., 0.5, 0.5)),
                        dtype=np.float32)
    pool_map = _dense_to_sparse(pool_map)
    data = np.reshape(np.arange(8).astype(np.float32), (4, 2))
    max_true = data[(1, 3), :]
    max_weighted = (data[(0, 2), :] + max_true) * 0.5

    pooled_max = gp.pool(data, pool_map, sizes=None, algorithm='max', name=None)
    pooled_weighted = gp.pool(
        data, pool_map, sizes=None, algorithm='weighted', name=None)

    self.assertAllClose(pooled_max, max_true)
    self.assertAllClose(pooled_weighted, max_weighted)

  @parameterized.parameters((20, 10, 3), (2, 1, 1), (2, 5, 4), (2, 1, 3))
  def test_pool_random(self, num_input_vertices, num_output_vertices,
                       num_features):
    """Tests pooling with random inputs."""
    pool_map = 0.001 + np.random.uniform(
        size=(num_output_vertices, num_input_vertices))
    data = np.random.uniform(size=(num_input_vertices, num_features))
    true_weighted = np.matmul(pool_map, data)
    true_max = np.tile(
        np.max(data, axis=0, keepdims=True), (num_output_vertices, 1))
    pool_map = _dense_to_sparse(pool_map)

    with self.subTest(name='max'):
      pooled_max = gp.pool(data, pool_map, None, algorithm='max')
      self.assertAllClose(pooled_max, true_max)

    with self.subTest(name='weighted'):
      pooled_weighted = gp.pool(data, pool_map, None, algorithm='weighted')
      self.assertAllClose(pooled_weighted, true_weighted)

  def test_pool_jacobian(self):
    """Tests the jacobian is correct."""
    sizes = ((2, 4), (3, 5))
    data_init = np.random.uniform(size=(2, 5, 3))
    pool_map = np.random.uniform(size=(2, 3, 5))
    data_init[0, -1, :] = 0.
    pool_map[0, -1, :] = 0.
    pool_map = _dense_to_sparse(pool_map)

    def gp_pool(data, algorithm):
      return gp.pool(data, pool_map, sizes, algorithm=algorithm)

    with self.subTest(name='max'):
      self.assert_jacobian_is_correct_fn(lambda data: gp_pool(data, 'max'),
                                         [data_init])

    with self.subTest(name='weighted'):
      self.assert_jacobian_is_correct_fn(lambda data: gp_pool(data, 'weighted'),
                                         [data_init])


class GraphPoolingTestUnpoolTests(test_case.TestCase):

  @parameterized.parameters(
      ("'sizes' must have an integer type.", np.float32, np.float32,
       np.float32),
      ("'data' must have a float type.", np.int32, np.float32, np.int32),
      ("'pool_map' and 'data' must have the same type.", np.float32, np.float64,
       np.int32))
  def test_unpool_exception_raised_types(self, err_msg, data_type,
                                         pool_map_type, sizes_type):
    """Tests the correct exceptions are raised for invalid types."""
    data = np.ones((2, 3, 3), dtype=data_type)
    pool_map = _dense_to_sparse(np.ones((2, 3, 3), dtype=pool_map_type))
    sizes = np.array(((1, 2), (2, 3)), dtype=sizes_type)

    with self.assertRaisesRegexp(TypeError, err_msg):
      gp.unpool(data, pool_map, sizes)

  @parameterized.parameters(
      ('data must have a rank greater than 1', (3,), (3,), None),
      ('pool_map must have a rank of 2', (3, 3), (3,), None),
      ('sizes must have a rank of 3', (4, 5, 3, 2), (4, 5, 3, 3), (3, 2)),
      ('data must have a rank less than 6', (2, 3, 4, 5, 3, 2),
       (2, 3, 4, 5, 3, 3), None),
  )
  def test_unpool_exception_raised_shapes(self, err_msg, data_shape,
                                          pool_map_shape, sizes_shape):
    """Tests the correct exceptions are raised for invalid shapes."""
    data = np.ones(data_shape, dtype=np.float32)
    pool_map = _dense_to_sparse(np.ones(pool_map_shape, dtype=np.float32))
    if sizes_shape is not None:
      sizes = np.ones(sizes_shape, dtype=np.int32)
    else:
      sizes = None

    with self.assertRaisesRegexp(ValueError, err_msg):
      gp.unpool(data, pool_map, sizes)

  @parameterized.parameters(
      ((2, 3), 4, 3, np.float32),
      ((1,), 6, 1, np.float32),
      ((4, 1, 3), 9, 7, np.float64),
      ((2, 8, 4), 19, 11, np.float64),
  )
  def test_unpool_identity(self, batch_shape, num_vertices, num_features,
                           data_type):
    """Tests graph unpooling with identity maps."""
    data_shape = np.concatenate((batch_shape, (num_vertices, num_features)))
    data = np.random.uniform(size=data_shape).astype(data_type)
    pool_map = _batch_sparse_eye(batch_shape, num_vertices, data_type)

    unpooled = gp.unpool(data, pool_map, sizes=None)
    self.assertAllClose(unpooled, data)

  def test_unpool_preset_padded(self):
    """Tests pooling with preset data and padding."""
    data = np.reshape(np.arange(12).astype(np.float32), (2, 3, 2))
    data[0, -1, :] = 0.
    sizes = ((2, 3), (3, 3))
    pool_map = _dense_to_sparse(
        np.array((((0.5, 0.5, 0.), (0., 0., 1.), (0., 0., 0.)),
                  ((1., 0., 0.), (0., 1., 0.), (0., 0., 1.))),
                 dtype=np.float32))

    unpooled = gp.unpool(data, pool_map, sizes)

    true = (((0., 1.), (0., 1.), (2., 3.)), ((6., 7.), (8., 9.), (10., 11.)))
    self.assertAllClose(unpooled, true)

  @parameterized.parameters((20, 4), (2, 1), (12, 4), (6, 3))
  def test_unpool_random(self, num_vertices, num_features):
    """Tests pooling with random data inputs."""
    output_vertices = num_vertices // 2
    pool_map = np.zeros(shape=(output_vertices, num_vertices), dtype=np.float32)
    for i in range(output_vertices):
      pool_map[i, (i * 2, i * 2 + 1)] = (0.5, 0.5)
    data = np.random.uniform(size=(output_vertices,
                                   num_features)).astype(np.float32)

    unpooled = gp.unpool(
        data, _dense_to_sparse(pool_map), sizes=None, name=None)

    with self.subTest(name='direct_unpool'):
      true = np.zeros(shape=(num_vertices, num_features)).astype(np.float32)
      true[0::2, :] = data
      true[1::2, :] = data
      self.assertAllClose(unpooled, true)

    with self.subTest(name='permute_pool_map'):
      permutation = np.random.permutation(num_vertices)
      pool_map_permute = pool_map[:, permutation]
      unpooled_permute = gp.unpool(data, _dense_to_sparse(pool_map_permute),
                                   None)
      true_permute = true[permutation, :]
      self.assertAllClose(unpooled_permute, true_permute)

  def test_unpool_jacobian_random(self):
    """Tests the jacobian is correct."""
    sizes = ((2, 4), (3, 5))
    data_init = np.random.uniform(size=(2, 3, 6))
    pool_map = np.random.uniform(size=(2, 3, 5))
    data_init[0, -1, :] = 0.
    pool_map[0, -1, :] = 0.
    pool_map = _dense_to_sparse(pool_map)

    def gp_unpool(data):
      return gp.unpool(data, pool_map, sizes)

    self.assert_jacobian_is_correct_fn(gp_unpool, [data_init])


class GraphPoolingUpsampleTransposeConvolutionTests(test_case.TestCase):

  @parameterized.parameters(
      ("'sizes' must have an integer type.", np.float32, np.float32,
       np.float32),
      ("'data' must have a float type.", np.int32, np.float32, np.int32),
      ("'pool_map' and 'data' must have the same type.", np.float32, np.float64,
       np.int32))
  def test_upsample_transposed_convolution_exception_raised_types(
      self, err_msg, data_type, pool_map_type, sizes_type):
    """Tests the correct exceptions are raised for invalid types."""
    data = np.ones((2, 3, 3), dtype=data_type)
    pool_map = _dense_to_sparse(np.ones((2, 3, 3), dtype=pool_map_type))
    sizes = np.array(((1, 2), (2, 3)), dtype=sizes_type)

    with self.assertRaisesRegexp(TypeError, err_msg):
      gp.upsample_transposed_convolution(
          data, pool_map, sizes, kernel_size=1, transposed_convolution_op=None)

  @parameterized.parameters(
      ('data must have a rank greater than 1', (3,), (3,), None),
      ('pool_map must have a rank of 2', (3, 3), (3,), None),
      ('sizes must have a rank of 3', (4, 5, 3, 2), (4, 5, 3, 3), (3, 2)),
      ('data must have a rank less than 6', (2, 3, 4, 5, 3, 2),
       (2, 3, 4, 5, 3, 3), None),
  )
  def test_upsample_transposed_convolution_exception_raised_shapes(
      self, err_msg, data_shape, pool_map_shape, sizes_shape):
    """Tests the correct exceptions are raised for invalid shapes."""
    data = np.ones(data_shape, dtype=np.float32)
    pool_map = _dense_to_sparse(np.ones(pool_map_shape, dtype=np.float32))
    if sizes_shape is not None:
      sizes = np.ones(sizes_shape, dtype=np.int32)
    else:
      sizes = None

    with self.assertRaisesRegexp(ValueError, err_msg):
      gp.upsample_transposed_convolution(
          data, pool_map, sizes, kernel_size=1, transposed_convolution_op=None)

  def test_upsample_transposed_convolution_exception_raised_callable(self):
    """Tests the correct exception is raised for a invalid convolution op."""
    data = np.ones((5, 3))
    pool_map = _dense_to_sparse(np.eye(5))
    err_msg = "'transposed_convolution_op' must be callable."

    with self.assertRaisesRegexp(TypeError, err_msg):
      gp.upsample_transposed_convolution(
          data,
          pool_map,
          sizes=None,
          kernel_size=1,
          transposed_convolution_op=1)

  @parameterized.parameters((1, 1, 1, np.float32), (5, 3, 1, np.float32),
                            (3, 6, 15, np.float64))
  def test_upsample_transposed_convolution_zero_kernel(self, num_vertices,
                                                       num_features,
                                                       kernel_size, data_type):
    """Tests the upsampling with a zero kernel."""
    data = np.random.uniform(size=(num_vertices,
                                   num_features)).astype(data_type)
    pool_map = np.zeros(
        shape=(num_vertices, num_vertices * kernel_size), dtype=data_type)
    for i in range(num_vertices):
      pool_map[i, np.arange(kernel_size * i, kernel_size *
                            (i + 1))] = (1.0 / kernel_size)
    pool_map = _dense_to_sparse(pool_map)
    # Transposed convolution op with a zero kernel.
    transposed_convolution_op = tf.keras.layers.Conv2DTranspose(
        filters=num_features,
        kernel_size=(1, kernel_size),
        strides=(1, kernel_size),
        padding='valid',
        use_bias=False,
        kernel_initializer=tf.compat.v1.keras.initializers.zeros())

    upsampled = gp.upsample_transposed_convolution(
        data,
        pool_map,
        sizes=None,
        kernel_size=kernel_size,
        transposed_convolution_op=transposed_convolution_op)

    # Initializes variables of the transpose conv layer.
    self.evaluate(tf.compat.v1.global_variables_initializer())

    self.assertAllEqual(
        tf.shape(input=upsampled), (num_vertices * kernel_size, num_features))
    self.assertAllEqual(upsampled, tf.zeros_like(upsampled))

  @parameterized.parameters(
      itertools.product((3,), (6,), (3,), range(3), range(6), range(6)),)
  def test_upsample_transposed_convolution_selector_kernel_random(
      self, num_vertices, num_features, kernel_size, kernel_index,
      feature1_index, feature2_index):
    """Tests the upsampling with an indicator kernel."""
    data = np.random.uniform(size=(num_vertices,
                                   num_features)).astype(np.float32)
    pool_map = np.zeros(
        shape=(num_vertices, num_vertices * kernel_size), dtype=np.float32)
    for i in range(num_vertices):
      pool_map[i, np.arange(kernel_size * i, kernel_size *
                            (i + 1))] = (1.0 / kernel_size)
    pool_map = _dense_to_sparse(pool_map)

    selection = np.zeros(
        shape=(1, kernel_size, num_features, num_features), dtype=np.float32)
    selection[0, kernel_index, feature1_index, feature2_index] = 1.
    initializer = tf.compat.v1.constant_initializer(value=selection)
    transposed_convolution_op = tf.keras.layers.Conv2DTranspose(
        filters=num_features,
        kernel_size=(1, kernel_size),
        strides=(1, kernel_size),
        padding='valid',
        use_bias=False,
        kernel_initializer=initializer)

    true = np.zeros(
        shape=(num_vertices * kernel_size, num_features), dtype=np.float32)
    input_column = feature2_index
    output_column = feature1_index
    output_row_start = kernel_index
    true[output_row_start::kernel_size, output_column] = (data[:, input_column])
    upsampled = gp.upsample_transposed_convolution(
        data,
        pool_map,
        sizes=None,
        kernel_size=kernel_size,
        transposed_convolution_op=transposed_convolution_op)

    # Initializes variables of the transpose conv layer.
    self.evaluate(tf.compat.v1.global_variables_initializer())

    self.assertAllEqual(upsampled, true)

  def test_upsample_transposed_convolution_preset_padded(self):
    """Tests upsampling with presets."""
    data = np.reshape(np.arange(12).astype(np.float32), (2, 3, 2))
    data[0, -1, :] = 0.
    sizes = ((2, 3), (3, 3))
    pool_map = _dense_to_sparse(
        np.array((((0.5, 0.5, 0.), (0., 0., 1.), (0., 0., 0.)),
                  ((1., 0., 0.), (0., 1., 0.), (0., 0., 1.))),
                 dtype=np.float32))

    kernel = np.ones(shape=(1, 2, 2, 2), dtype=np.float32)
    initializer = tf.compat.v1.constant_initializer(value=kernel)
    transposed_convolution_op = tf.keras.layers.Conv2DTranspose(
        filters=2,
        kernel_size=(1, 2),
        strides=(1, 2),
        padding='valid',
        use_bias=False,
        kernel_initializer=initializer)

    # Convolving with an all-ones kernel is equal to summation of the input.
    data_sum = np.tile(np.sum(data, axis=-1, keepdims=True), (1, 1, 2))
    true = np.zeros(shape=(2, 3, 2), dtype=np.float32)
    true[0, :, :] = data_sum[0, (0, 0, 1), :]
    true[1, :, :] = data_sum[1, :, :]
    upsampled = gp.upsample_transposed_convolution(
        data,
        pool_map,
        sizes=sizes,
        kernel_size=2,
        transposed_convolution_op=transposed_convolution_op)

    # Initializes variables of the transpose conv layer.
    self.evaluate(tf.compat.v1.global_variables_initializer())

    self.assertAllEqual(upsampled.shape, (2, 3, 2))
    self.assertAllClose(upsampled, true)

  def test_upsample_transposed_convolution_jacobian_random(self):
    """Tests the jacobian is correct."""
    num_filters = 6
    kernel_size = 1
    data_init = np.random.uniform(size=(2, 5, num_filters))
    pool_map = _batch_sparse_eye((2,), 5, np.float64)
    transposed_convolution_op = tf.keras.layers.Conv2DTranspose(
        filters=num_filters,
        kernel_size=(1, kernel_size),
        strides=(1, kernel_size),
        padding='valid',
        dtype='float64')

    # Calling the upsample_transposed_convolution to create the variables
    # in the transposed_convoution.
    gp.upsample_transposed_convolution(
        data_init,
        pool_map,
        sizes=None,
        kernel_size=kernel_size,
        transposed_convolution_op=transposed_convolution_op)
    # Initializes variables of the transpose conv layer.
    self.evaluate(tf.compat.v1.global_variables_initializer())

    def gp_upsample_transposed_convolution(data):
      return gp.upsample_transposed_convolution(
          data,
          pool_map,
          sizes=None,
          kernel_size=kernel_size,
          transposed_convolution_op=transposed_convolution_op)

    self.assert_jacobian_is_correct_fn(gp_upsample_transposed_convolution,
                                       [data_init])

  def test_upsample_transposed_convolution_jacobian_random_padding(self):
    """Tests the jacobian is correct with padded data."""
    num_filters = 6
    sizes = ((2, 4), (3, 5))
    data_init = np.random.uniform(size=(2, 3, num_filters))
    data_init[0, -1, :] = 0.
    pool_map = np.array(
        (((0.5, 0.5, 0., 0., 0.), (0., 0., 0.5, 0.5, 0.), (0., 0., 0., 0., 0.)),
         ((1., 0., 0., 0., 0.), (0., 1. / 3., 1. / 3., 1. / 3., 0.),
          (0., 0., 0., 0., 1.))),
        dtype=data_init.dtype)
    pool_map = _dense_to_sparse(pool_map)
    kernel_size = 2
    transposed_convolution_op = tf.keras.layers.Conv2DTranspose(
        filters=num_filters,
        kernel_size=(1, kernel_size),
        strides=(1, kernel_size),
        padding='valid',
        dtype='float64')

    # Calling the upsample_transposed_convolution to create the variables
    # in the transposed_convoution.
    gp.upsample_transposed_convolution(
        data_init,
        pool_map,
        sizes=sizes,
        kernel_size=kernel_size,
        transposed_convolution_op=transposed_convolution_op)
    # Initializes variables of the transpose conv layer.
    self.evaluate(tf.compat.v1.global_variables_initializer())

    def gp_upsample_transposed_convolution(data):
      return gp.upsample_transposed_convolution(
          data,
          pool_map,
          sizes=sizes,
          kernel_size=kernel_size,
          transposed_convolution_op=transposed_convolution_op)

    self.assert_jacobian_is_correct_fn(gp_upsample_transposed_convolution,
                                       [data_init])


if __name__ == '__main__':
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements graph convolutions layers."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

import tensorflow_graphics.geometry.convolution.graph_convolution as gc
from tensorflow_graphics.util import export_api


def feature_steered_convolution_layer(
    data,
    neighbors,
    sizes,
    translation_invariant=True,
    num_weight_matrices=8,
    num_output_channels=None,
    initializer=tf.compat.v1.truncated_normal_initializer(stddev=0.1),
    name=None,
    var_name=None):
  # pyformat: disable
  """Wraps the function `feature_steered_convolution` as a TensorFlow layer.

  The shorthands used below are
    `V`: The number of vertices.
    `C`: The number of channels in the input data.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    data: A `float` tensor with shape `[A1, ..., An, V, C]`.
    neighbors: A SparseTensor with the same type as `data` and with shape
      `[A1, ..., An, V, V]` representing vertex neighborhoods. The neighborhood
      of a vertex defines the support region for convolution. For a mesh, a
      common choice for the neighborhood of vertex `i` would be the vertices in
      the K-ring of `i` (including `i` itself). Each vertex must have at least
      one neighbor. For a faithful implementation of the FeaStNet paper,
      neighbors should be a row-normalized weight matrix corresponding to the
      graph adjacency matrix with self-edges:
      `neighbors[A1, ..., An, i, j] > 0` if vertex `i` and `j` are neighbors,
      `neighbors[A1, ..., An, i, i] > 0` for all `i`, and
      `sum(neighbors, axis=-1)[A1, ..., An, i] == 1.0` for all `i`.
      These requirements are relaxed in this implementation.
    sizes: An `int` tensor of shape `[A1, ..., An]` indicating the true input
      sizes in case of padding (`sizes=None` indicates no padding).
      `sizes[A1, ..., An] <= V`. If `data` and `neighbors` are 2-D, `sizes` will
      be ignored. As an example, consider an input consisting of three graphs
      `G0`, `G1`, and `G2` with `V0`, `V1` and `V2` vertices respectively. The
      padded input would have the following shapes: `data.shape = [3, V, C]`,
      and `neighbors.shape = [3, V, V]`, where `V = max([V0, V1, V2])`. The true
      sizes of each graph will be specified by `sizes=[V0, V1, V2]`.
      `data[i, :Vi, :]` and `neighbors[i, :Vi, :Vi]` will be the vertex and
      neighborhood data of graph `Gi`. The `SparseTensor` `neighbors` should
      have no nonzero entries in the padded regions.
    translation_invariant: A `bool`. If `True` the assignment of features to
      weight matrices will be invariant to translation.
    num_weight_matrices: An `int` specifying the number of weight matrices used
      in the convolution.
    num_output_channels: An optional `int` specifying the number of channels in
      the output. If `None` then `num_output_channels = C`.
    initializer: An initializer for the trainable variables.
    name: A (name_scope) name for this op. Passed through to
      feature_steered_convolution().
    var_name: A (var_scope) name for the variables. Defaults to
      `graph_convolution_feature_steered_convolution_weights`.

  Returns:
    Tensor with shape `[A1, ..., An, V, num_output_channels]`.
  """
  # pyformat: enable
  with tf.compat.v1.variable_scope(
      var_name,
      default_name='graph_convolution_feature_steered_convolution_weights'):
    # Skips shape validation to avoid redundancy with
    # feature_steered_convolution().
    data = tf.convert_to_tensor(value=data)

    in_channels = tf.compat.v1.dimension_value(data.shape[-1])
    if num_output_channels is None:
      out_channels = in_channels
    else:
      out_channels = num_output_channels

    var_u = tf.compat.v1.get_variable(
        shape=(in_channels, num_weight_matrices),
        dtype=data.dtype,
        initializer=initializer,
        name='u')
    if translation_invariant:
      var_v = -var_u
    else:
      var_v = tf.compat.v1.get_variable(
          shape=(in_channels, num_weight_matrices),
          dtype=data.dtype,
          initializer=initializer,
          name='v')
    var_c = tf.compat.v1.get_variable(
        shape=(num_weight_matrices),
        dtype=data.dtype,
        initializer=initializer,
        name='c')
    var_w = tf.compat.v1.get_variable(
        shape=(num_weight_matrices, in_channels, out_channels),
        dtype=data.dtype,
        initializer=initializer,
        name='w')
    var_b = tf.compat.v1.get_variable(
        shape=(out_channels),
        dtype=data.dtype,
        initializer=initializer,
        name='b')

    return gc.feature_steered_convolution(
        data=data,
        neighbors=neighbors,
        sizes=sizes,
        var_u=var_u,
        var_v=var_v,
        var_c=var_c,
        var_w=var_w,
        var_b=var_b,
        name=name)


class FeatureSteeredConvolutionKerasLayer(tf.keras.layers.Layer):
  """Wraps the function `feature_steered_convolution` as a Keras layer."""

  def __init__(self,
               translation_invariant=True,
               num_weight_matrices=8,
               num_output_channels=None,
               initializer=None,
               name=None,
               **kwargs):
    """Initializes FeatureSteeredConvolutionKerasLayer.

    Args:
      translation_invariant: A `bool`. If `True` the assignment of features to
        weight matrices will be invariant to translation.
      num_weight_matrices: An `int` specifying the number of weight matrices
        used in the convolution.
      num_output_channels: An optional `int` specifying the number of channels
        in the output. If `None` then `num_output_channels` will be the same as
        the input dimensionality.
      initializer: An initializer for the trainable variables. If `None`,
        defaults to `tf.compat.v1.truncated_normal_initializer(stddev=0.1)`.
      name: A name for this layer.
      **kwargs: Additional keyword arguments passed to the base layer.
    """
    super(FeatureSteeredConvolutionKerasLayer, self).__init__(
        name=name, **kwargs)
    self._num_weight_matrices = num_weight_matrices
    self._num_output_channels = num_output_channels
    self._translation_invariant = translation_invariant
    if initializer is None:
      self._initializer = tf.compat.v1.truncated_normal_initializer(stddev=0.1)
    else:
      self._initializer = initializer

  def build(self, input_shape):
    """Initializes the trainable weights."""
    in_channels = tf.TensorShape(input_shape[0]).as_list()[-1]
    if self._num_output_channels is None:
      out_channels = in_channels
    else:
      out_channels = self._num_output_channels
    dtype = self.dtype
    num_weight_matrices = self._num_weight_matrices
    initializer = self._initializer

    self.var_u = self.add_weight(
        shape=(in_channels, num_weight_matrices),
        dtype=dtype,
        initializer=initializer,
        name='u')
    if self._translation_invariant:
      self.var_v = -self.var_u
    else:
      self.var_v = self.add_weight(
          shape=(in_channels, num_weight_matrices),
          dtype=dtype,
          initializer=initializer,
          name='v')
    self.var_c = self.add_weight(
        shape=(num_weight_matrices,),
        dtype=dtype,
        initializer=initializer,
        name='c')
    self.var_w = self.add_weight(
        shape=(num_weight_matrices, in_channels, out_channels),
        dtype=dtype,
        initializer=initializer,
        name='w',
        trainable=True)
    self.var_b = self.add_weight(
        shape=(out_channels,),
        dtype=dtype,
        initializer=initializer,
        name='b',
        trainable=True)

  def call(self, inputs, **kwargs):
    # pyformat: disable
    """Executes the convolution.

    The shorthands used below are
      `V`: The number of vertices.
      `C`: The number of channels in the input data.

    Note:
      In the following, A1 to An are optional batch dimensions.

    Args:
      inputs: A list of two tensors `[data, neighbors]`. `data` is a `float`
        tensor with shape `[A1, ..., An, V, C]`. `neighbors` is a `SparseTensor`
        with the same type as `data` and with shape `[A1, ..., An, V, V]`
        representing vertex neighborhoods. The neighborhood of a vertex defines
        the support region for convolution. For a mesh, a common choice for the
        neighborhood of vertex `i` would be the vertices in the K-ring of `i`
        (including `i` itself). Each vertex must have at least one neighbor. For
        a faithful implementation of the FeaStNet paper, neighbors should be a
        row-normalized weight matrix corresponding to the graph adjacency matrix
        with self-edges: `neighbors[A1, ..., An, i, j] > 0` if vertex `j` is a
        neighbor of vertex `i`, and `neighbors[A1, ..., An, i, i] > 0` for all
        `i`, and `sum(neighbors, axis=-1)[A1, ..., An, i] == 1.0` for all `i`.
        These requirements are relaxed in this implementation.
      **kwargs: A dictionary containing the key `sizes`, which is an `int` tensor
        of shape `[A1, ..., An]` indicating the true input sizes in case of
        padding (`sizes=None` indicates no padding). `sizes[A1, ..., An] <= V`.
        If `data` and `neighbors` are 2-D, `sizes` will be ignored. As an
        example usage of `sizes`, consider an input consisting of three graphs
        `G0`, `G1`, and `G2` with `V0`, `V1`, and `V2` vertices respectively.
        The padded input would have the shapes `data.shape = [3, V, C]`, and
        `neighbors.shape = [3, V, V]`, where `V = max([V0, V1, V2])`. The true
        sizes of each graph will be specified by `sizes=[V0, V1, V2]`.
        `data[i, :Vi, :]` and `neighbors[i, :Vi, :Vi]` will be the vertex and
        neighborhood data of graph `Gi`. The `SparseTensor` `neighbors` should
        have no nonzero entries in the padded regions.

    Returns:
      Tensor with shape `[A1, ..., An, V, num_output_channels]`.
    """

    # pyformat: enable
    sizes = kwargs.get('sizes', None)
    return gc.feature_steered_convolution(
        data=inputs[0],
        neighbors=inputs[1],
        sizes=sizes,
        var_u=self.var_u,
        var_v=self.var_v,
        var_c=self.var_c,
        var_w=self.var_w,
        var_b=self.var_b)


class DynamicGraphConvolutionKerasLayer(tf.keras.layers.Layer):
  """A keras layer for dynamic graph convolutions.

  Dynamic GraphCNN for Learning on Point Clouds
  Yue Wang, Yongbin Sun, Ziwei Liu, Sanjay E. Sarma, Michael Bronstein, and
  Justin Solomon.
  https://arxiv.org/abs/1801.07829

  This layer implements an instance of the graph convolutional operation
  described in the paper above, specifically a graph convolution block with a
  single edge filtering layer. This implementation is intended to demonstrate
  how `graph_convolution.edge_convolution_template` can be wrapped to implement
  a variety of edge convolutional methods.

  This implementation is slightly generalized version to what is described in
  the paper in that here variable sized neighborhoods are allowed rather than
  forcing a fixed size k-neighbors. Users must provide the neighborhoods as
  input.
  """

  def __init__(self,
               num_output_channels,
               reduction,
               activation=None,
               use_bias=True,
               kernel_initializer='glorot_uniform',
               bias_initializer='zeros',
               kernel_regularizer=None,
               bias_regularizer=None,
               activity_regularizer=None,
               kernel_constraint=None,
               bias_constraint=None,
               name=None,
               **kwargs):
    """Initializes DynamicGraphConvolutionKerasLayer.

    Args:
      num_output_channels: An `int` specifying the number of output channels.
      reduction: Either 'weighted' or 'max'. Specifies the reduction over
        neighborhood edge features as described in the paper above.
      activation: The `activation` argument of `tf.keras.layers.Conv1D`.
      use_bias: The `use_bias` argument of `tf.keras.layers.Conv1D`.
      kernel_initializer: The `kernel_initializer` argument of
        `tf.keras.layers.Conv1D`.
      bias_initializer: The `bias_initializer` argument of
        `tf.keras.layers.Conv1D`.
      kernel_regularizer: The `kernel_regularizer` argument of
        `tf.keras.layers.Conv1D`.
      bias_regularizer: The `bias_regularizer` argument of
        `tf.keras.layers.Conv1D`.
      activity_regularizer: The `activity_regularizer` argument of
        `tf.keras.layers.Conv1D`.
      kernel_constraint: The `kernel_constraint` argument of
        `tf.keras.layers.Conv1D`.
      bias_constraint: The `bias_constraint` argument of
        `tf.keras.layers.Conv1D`.
      name: A name for this layer.
      **kwargs: Additional keyword arguments passed to the base layer.
    """
    super(DynamicGraphConvolutionKerasLayer, self).__init__(
        name=name, **kwargs)
    self._num_output_channels = num_output_channels
    self._reduction = reduction
    self._activation = activation
    self._use_bias = use_bias
    self._kernel_initializer = kernel_initializer
    self._bias_initializer = bias_initializer
    self._kernel_regularizer = kernel_regularizer
    self._bias_regularizer = bias_regularizer
    self._activity_regularizer = activity_regularizer
    self._kernel_constraint = kernel_constraint
    self._bias_constraint = bias_constraint

  def build(self, input_shape):  # pylint: disable=unused-argument
    """Initializes the layer weights."""
    self._conv1d_layer = tf.keras.layers.Conv1D(
        filters=self._num_output_channels,
        kernel_size=1,
        strides=1,
        padding='valid',
        activation=self._activation,
        use_bias=self._use_bias,
        kernel_initializer=self._kernel_initializer,
        bias_initializer=self._bias_initializer,
        kernel_regularizer=self._kernel_regularizer,
        bias_regularizer=self._bias_regularizer,
        activity_regularizer=self._activity_regularizer,
        kernel_constraint=self._kernel_constraint,
        bias_constraint=self._bias_constraint)

  def call(self, inputs, **kwargs):
    # pyformat: disable
    """Executes the convolution.

    The shorthands used below are
      `V`: The number of vertices.
      `C`: The number of channels in the input data.

    Note:
      In the following, A1 to An are optional batch dimensions.

    Args:
      inputs: A list of two tensors `[data, neighbors]`. `data` is a `float`
        tensor with shape `[A1, ..., An, V, C]`. `neighbors` is a `SparseTensor`
        with the same type as `data` and with shape `[A1, ..., An, V, V]`
        representing vertex neighborhoods. The neighborhood of a vertex defines
        the support region for convolution. For a mesh, a common choice for the
        neighborhood of vertex `i` would be the vertices in the K-ring of `i`
        (including `i` itself). Each vertex must have at least one neighbor. For
        `reduction='weighted'`, `neighbors` should be a row-normalized matrix:
        `sum(neighbors, axis=-1)[A1, ..., An, i] == 1.0` for all `i`, although
        this is not enforced in the implementation in case different neighbor
        weighting schemes are desired.
      **kwargs: A dictionary containing the key `sizes`, which is an `int`
        tensor of shape `[A1, ..., An]` indicating the true input sizes in case
        of padding (`sizes=None` indicates no padding).
        `sizes[A1, ..., An] <= V`. If `data` and `neighbors` are 2-D, `sizes`
        will be ignored. As an example usage of `sizes`, consider an input
        consisting of three graphs `G0`, `G1`, and `G2` with `V0`, `V1`, and
        `V2` vertices respectively. The padded input would have the shapes
        `data.shape = [3, V, C]`, and `neighbors.shape = [3, V, V]`,
        where `V = max([V0, V1, V2])`. The true sizes of each graph will be
        specified by `sizes=[V0, V1, V2]`. `data[i, :Vi, :]` and
        `neighbors[i, :Vi, :Vi]` will be the vertex and neighborhood data of
        graph `Gi`. The `SparseTensor` `neighbors` should have no nonzero
        entries in the padded regions.

    Returns:
      Tensor with shape `[A1, ..., An, V, num_output_channels]`.
    """
    # pyformat: enable
    def _edge_convolution(vertices, neighbors, conv1d_layer):
      r"""The edge filtering op passed to `edge_convolution_template`.

      This instance implements the edge function
      $$h_{\theta}(x, y) = MLP_{\theta}([x, y - x])$$

      Args:
        vertices: A 2-D Tensor with shape `[D1, D2]`.
        neighbors: A 2-D Tensor with the same shape and type as `vertices`.
        conv1d_layer: A callable 1d convolution layer.

      Returns:
        A 2-D Tensor with shape `[D1, D3]`.
      """
      concat_features = tf.concat(
          values=[vertices, neighbors - vertices], axis=-1)
      concat_features = tf.expand_dims(concat_features, 0)
      convolved_features = conv1d_layer(concat_features)
      convolved_features = tf.squeeze(input=convolved_features, axis=(0,))
      return convolved_features

    sizes = kwargs.get('sizes', None)
    return gc.edge_convolution_template(
        data=inputs[0],
        neighbors=inputs[1],
        sizes=sizes,
        edge_function=_edge_convolution,
        reduction=self._reduction,
        edge_function_kwargs={'conv1d_layer': self._conv1d_layer})


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()
